// *************************************************************************
//    Term.h
// *************************************************************************

#ifndef _Term_H_
#define _Term_H_

#include "CArg.h"
#include "Entity.h"
#include "Debug.h"

// *************************************************************************
//
// A Term is implemented using a variant of structure sharing
// (though this may change):
// The 'expr' field contains the static structure of the term,
// while 'bindenv' maps (some of) the variables in 'expr' to terms.
// Variables in these `terms' are not interpreted in BindEnv, unlike what is
// done in Boyer and Moore structure sharing.
//

class Cor_QTerm;

class Cor_Term {

  public:
      // arg is NULL if unbound???
    Cor_BindEnv *bindenv;     // context for variables in expr

      // Note that there is 4 bytes wasted here because of packing
    //unsigned int PACK;

    Cor_CArg     expr;

#ifdef    REUSE_MEMORY
    void* operator new(size_t size) {
        return(Coral_Free->acquire(size));
    }

    void  operator delete(void *trash, size_t size) {
        Coral_Free->dispose((Cor_Ent_Element*) trash, size);
    }
#endif

    Cor_Term() { bindenv = NULL; }

    Cor_Term (Cor_CArg *arg1, Cor_BindEnv *bindenv1 = NULL){
        bindenv = bindenv1;
        if (arg1)
            expr = *arg1;
    }

    Cor_Term(const Cor_CArg& _expr, Cor_BindEnv *env = NULL)
                : expr(_expr) {
        bindenv = env;
    }

// ~Cor_Term(){};

    inline Cor_Term& operator= (const Cor_Term& term) {
        bindenv = term.bindenv;
        expr = term.expr;
        return(*this);
    }

    inline void set(Cor_CArg& _expr, Cor_BindEnv *env = NULL) {
        expr = _expr;
        bindenv = env;
    }

    inline void clear() {
        bindenv = NULL;
        expr.free_mem();
    }

    inline void printon(ostream *file) const {
        if (!expr.is_empty())
            expr.print(bindenv, file);
        else
            *file << '_';
    }

      // Stream printing
    friend ostream& operator<< (ostream& out, const Cor_Term& term);

      // Assuming 'expr' is a variable, get its bindings.
    inline Cor_Term dereference_var() const;
    inline Variable varnum() { return expr.var_num(); }
    Cor_Term dereference();

    inline int hash() const { return expr.hash(bindenv); }

    inline CArgKind Kind()   { return expr.Kind(); }

    inline int isConstant() { return(expr.isConstant()); }
};

// *************************************************************************
// The same as above, but "quick".  Since this version uses a pointer
// instead of a copy of the CArg, no changes should be made to it, and
// thus some methods have been eliminated.

class Cor_QTerm {

  public:
      // arg is NULL if unbound???
    Cor_BindEnv *bindenv;     // context for variables in expr

    Cor_CArg    *expr;

#ifdef    REUSE_MEMORY
    void* operator new(size_t size) {
        return(Coral_Free->acquire(size));
    }

    void  operator delete(void *trash, size_t size) {
        Coral_Free->dispose((Cor_Ent_Element*) trash, size);
    }
#endif

    Cor_QTerm (Cor_CArg *arg1 = NULL, Cor_BindEnv *bindenv1 = NULL) {
        bindenv = bindenv1;
        expr    = arg1;
    }

    Cor_QTerm(const Cor_CArg& _expr, Cor_BindEnv *env = NULL) {
        expr = (Cor_CArg*) &_expr;
        bindenv = env;
    }

    inline Cor_QTerm& operator= (const Cor_Term& term) {
        bindenv = term.bindenv;
        expr = (Cor_CArg*) &(term.expr);
        return(*this);
    }

    inline Cor_QTerm& operator= (const Cor_QTerm& term) {
        bindenv = term.bindenv;
        expr = term.expr;
        return(*this);
    }

// ~Cor_QTerm(){};

    inline void set(Cor_CArg& _expr, Cor_BindEnv *env = NULL) {
        expr    = &_expr;
        bindenv = env;
    }

    inline void clear() {
        bindenv = NULL;
        expr    = NULL;
    }

    inline void printon(ostream *file) const {
        if (expr && !expr->is_empty())
            expr->print(bindenv, file);
        else
            *file << '_';
    }

      // Stream printing
    friend ostream& operator<< (ostream& out, const Cor_QTerm& term);

      // Assuming 'expr' is a variable, get its bindings.
    inline Cor_Term dereference_var() const;
    inline Variable varnum() { return expr->var_num(); }
    Cor_Term dereference();

    inline int hash() const { return expr->hash(bindenv); }

    inline CArgKind Kind()   { return expr->Kind(); }

    inline int isConstant() { return(expr->isConstant()); }
};

// *************************************************************************
// TermLink is used by CopyArgs() to re-number variables

class Cor_TermLink;

class Cor_TermLink {
  public:
    Cor_TermLink *next;        // Next TermLink in list
    Variable      old_var;
    Cor_BindEnv  *old_bindenv;
    Cor_CArg      new_arg;

    void* operator new(size_t size) {
        return(Coral_Free->acquire(size));
    }

    void  operator delete(void *trash, size_t size) {
        Coral_Free->dispose((Cor_Ent_Element*) trash, size);
    }
};

// *************************************************************************

#endif _Term_H_
