/*****************************************************************************/
//	BitVector.C
/*****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include "BitVector.h"

/*****************************************************************************/
//	Number of BitWords to represent 'i' bits

static int bit_words_needed(int i)
{
    return (i + BIT_VECTOR_WORD - 1) >> BIT_VECTOR_WORD_LOG;
}

/*****************************************************************************/
void BitVector::init(const BitVector& b2)
{
    int words_needed = bit_words_needed(b2.length);
    if (words_needed <= 1) {
	u.word0 = b2.u.word0;
    } else {
	u.words = (BitWord*)malloc(words_needed * sizeof(BitWord));
	BitWord *src = b2.u.words;
	BitWord *dst = u.words;
	while (--words_needed >= 0) *dst++ = *src++;
    }
    length = b2.length;
}

/*****************************************************************************/
BitVector::~BitVector()
{
    if (length > BIT_VECTOR_WORD) free(u.words);
}

/*****************************************************************************/
void BitVector::printon(FILE *file) const
{
    for (int i = 0; i < length; ++i)
      putc('0' + test(i), file);
}

/*****************************************************************************/
int BitVector::next(int i) const
{
    for (++i; i < length; ++i)
      if (test(i))
	return i;
    return -1;
}

/*****************************************************************************/
void BitVector::setlen(int l)
{
    if (l > BIT_VECTOR_WORD) {
	BitWord *old_words = nc_words();
	int old_length = bit_words_needed(length);
	int new_length = bit_words_needed(l);
	register int j;
	int new_bytes = new_length * sizeof(BitWord);
	BitWord *new_words = (BitWord*)malloc(new_bytes);
	memset(new_words, 0, new_bytes);
	for (j = old_length; --j >= 0; ) new_words[j] = old_words[j];
	u.words = new_words;
	if (old_length > 1) free(old_words);
    } else if (length > BIT_VECTOR_WORD) {
	BitWord w = u.words[0];
	free(u.words);
	u.word0 = w;
    }
    length = l;
}

/*****************************************************************************/
void BitVector::set(int index, int value /*= 1*/)
{
    if (index >= length)
      setlen(index+1);
    BitWord *ptr = &nc_words()[index>>BIT_VECTOR_WORD_LOG];
    BitWord mask = 1 << (index & BIT_VECTOR_WORD-1);
    if (value)
      *ptr |= mask;
    else
      *ptr &= ~mask;
}

/*****************************************************************************/
void BitVector::operator=(const BitVector& b2)
{
    // Ok for (&b2 == this).
    int words_needed = bit_words_needed(b2.length);
    if (words_needed <= 1) {
	if (length > BIT_VECTOR_WORD)
	  free(u.words);
	u.word0 = b2.u.word0;
    } else {
	if (length > BIT_VECTOR_WORD)
	  u.words = (BitWord*)realloc(u.words, words_needed*sizeof(BitWord));
	else
	  u.words = (BitWord*)malloc(words_needed * sizeof(BitWord));
	BitWord *src = b2.u.words;
	BitWord *dst = u.words;
	while (--words_needed >= 0) *dst++ = *src++;
    }
    length = b2.length;
}

/*****************************************************************************/
int BitVector::operator==(const BitVector& b2) const
{
    if (length != b2.length) return 0;

    register const BitWord *ptr1 = words();
    register const BitWord *ptr2 = b2.words();
    int words1 = bit_words_needed(length);
    int words2 = bit_words_needed(b2.length);
    register int i = words1 > words2 ? words2 : words1;
    while (--i >= 0) {
	if (*ptr1++ != *ptr2++)
	  return 0;
    }
    i = words1 - words2;
    if (i > 0) {
	while (--i >= 0) if (*ptr1++ != 0) return 0;
    } else if (i < 0) {
	for (i = -1; --i >= 0; ) {
	    if (*ptr2++ != 0)
	      return 0;
	}
    }
    return 1;
}

/*****************************************************************************/
int BitVector::implies(const BitVector& b2) const
{
    register const BitWord *ptr1 = words();
    register const BitWord *ptr2 = b2.words();
    int words1 = bit_words_needed(length);
    int words2 = bit_words_needed(b2.length);
    register int i = words1 > words2 ? words2 : words1;

    while (--i >= 0) {
	if (*ptr1++ & ~*ptr2++)
	  return 0;
    }
    i = words1 - words2;
    while (--i >= 0) {
	if (*ptr1++ != 0)
	  return 0;
    }
    return 1;
}

/*****************************************************************************/
int BitVector::count() const
{
    int i, c;
    for(c = next(-1), i = 0; c >= 0; c = next(c) )
      i++;
    return(i) ;
}

/*****************************************************************************/
void BitVector::or_bits(const BitVector *b2)
{
    int c;
    for(c = b2->next(-1); c >= 0; c = b2->next(c) )
      set(c);
}

/*****************************************************************************/
void BitVector::clear()
{
    memset(nc_words(), 0, bit_words_needed(length) * sizeof(BitWord));
}

/*****************************************************************************/
int BitVector::any() const
{
    register const BitWord *ptr = words();
    register int i = bit_words_needed(length);
    while (--i >= 0) {
	if (*ptr++)
	  return 1;
    }
    return 0;
}

/*****************************************************************************/
int BitVector::operator&(const BitVector& b2) const
{
    register const BitWord *ptr1 = words();
    register const BitWord *ptr2 = b2.words();
    register int i = length;

    if (i > b2.length)
      i = b2.length;
    i = (i+BIT_VECTOR_WORD-1) >> BIT_VECTOR_WORD_LOG;
    while (--i >= 0) {
      if (*ptr1++ & *ptr2++)
	return 1;
  }
    return 0;
}

/*****************************************************************************/

