// *************************************************************************
//    BindTime.C
// *************************************************************************

#include <stdlib.h>
#include <string.h>

#include "BindTime.h"

#include "CArg.h"
#include "PredAnnotations.h"
#include "Globals.h"
#include "RuleInfo.h"
#include "Literal.h"
#include "RelationMisc.h"
#include "BuiltinSymbols.h"
#include "Relation.h"
#include "Rule.h"
#include "Debug.h"

#ifdef    BIND_VERBOSE
#include "Unparse.h"
#endif

// *************************************************************************

static void check_this_arg(Cor_CArg*, Cor_BitVector*);

// *************************************************************************
static void read_vars_from_arglist(Cor_BitVector *varvect, Cor_ArgList& args)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    FOR_EACH_ARG(a, args) {
        ASSERT((! a->is_empty()));
        check_this_arg(a, varvect);
    } END_EACH_ARG
}

// *************************************************************************
static void check_this_arg(Cor_CArg *a, Cor_BitVector *varvect)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch(a->Kind()) {

    case VAR_ARG:
        varvect->set(a->var_num());
        break;

    case FUNCT_ARG:
        read_vars_from_arglist(varvect, a->args());
        break;

    case LIST_ARG: {
        int j;
        for (j=a->num_items()-1; j >= 0; j--) {
            Cor_CArg *tmp = a->items()->get_ith(j);
            check_this_arg(tmp, varvect);
        }
        break;
    }

    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:
    case STDDEV_GRPING_ARG:
    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:
    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
	    read_vars_from_arglist(varvect, a->nested_args()->args());
        break;

    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:
    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:
    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        check_this_arg(a->larg(), varvect);
        check_this_arg(a->rarg(), varvect);
        break;

    default:
        break;
    }
}

// *************************************************************************
int Cor_BindTime::Analyze(Cor_RuleInfo *rInfo, int *predOrders)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j, k, tmp_num_vars;

      // Clear out any answers from previous analysis.
    if (first_bind)
        clear();

    num_lits = rInfo->num_literals;

        // The head of the rule
    num_vars = (rInfo->arg_list[num_lits]).max_vars();

    for (j=0; j < num_lits; j++) {
          // We don't care about the order, so don't need to use predOrders
        tmp_num_vars = (rInfo->arg_list[j]).max_vars();

          // Save the maximum variable number seen
        num_vars = (tmp_num_vars > num_vars) ? tmp_num_vars : num_vars;
    }

      // Create the structures needed
    head_vars = new Cor_BitVector(num_vars);

    first_bind = new int [num_vars];
    last_seen  = new int [num_vars];
    lit_kind   = new Cor_BT_Literal_Kind [num_lits];

    lit_vars   = new Cor_BitVector* [num_lits];

    for (j=0; j < num_lits; j++) {
        lit_kind[j]   = COR_BT_LIT_NORMAL;
        lit_vars[j]   = new Cor_BitVector(num_vars);
    }

    for (j=0; j < num_vars; j++) {
        first_bind[j] = num_lits;   // an invalid value
        last_seen[j]  = -1;
    }

      // Mark all BitVector's with the variables that occur there
    read_vars_from_arglist(head_vars, rInfo->arg_list[num_lits]);

    for (j=0; j < num_lits; j++) {
        k = (predOrders) ? predOrders[j] : j;

          // Mark cut's and single-success predicates
        if (rInfo->rule->preds[k]->name().equals(Cor_CutSymbol))
            lit_kind[k] = COR_BT_LIT_CUT;

        else if (NEGATED(k)) {
            lit_kind[k] = COR_BT_LIT_SINGLE;
        } else if (rInfo->rule->preds[k]->arith_pred) {
            lit_kind[k] = COR_BT_LIT_ARITHPRED;
        } else {
            Cor_Relation *rel;
            rel = Cor_find_external_relation( rInfo->rule->preds[k]->name(),
                    rInfo->rule->preds[k]->arity());

            if (rel && rel->single_success) {
                lit_kind[k] = COR_BT_LIT_SINGLE;
            }
        }

        read_vars_from_arglist(lit_vars[k], rInfo->arg_list[k]);
    }

      // Okay, we have all the data, now do the actual binding-analysis
    int m;

    for (j=0; j < num_lits; j++) {
        k = (predOrders) ? predOrders[j] : j;
        for (m = 0; m < num_vars; m++) {
            if (lit_vars[j]->test(m)) {
                if (k < first_bind[m])
                    first_bind[m] = k;
                if (k > last_seen[m])
                    last_seen[m] = k;
            }
        }
    }


#ifdef    BIND_VERBOSE
    cout << "-----------------" << endl;
    cout << *(rInfo->rule) << endl;
    show_raw_info(predOrders);
#endif

      // Check that arithmetic predicates are not binding variables
    int bind_ok = 1;
    for (k = 0; k < num_vars; k++) {

        if ((first_bind[k] < num_lits)
            && (lit_kind[first_bind[k]] == COR_BT_LIT_ARITHPRED)
                && (first_bind[k] != last_seen[k])) {
            bind_ok = 0;
        /*
            cout << "Warning arith predidicate as first binding for "
                 << "X_" << k << "_ at literal " << first_bind[k] << endl;
        */
            break;
        }
    }

    return(bind_ok);
}

// *************************************************************************
// Set the functions to use when unifying later
int Cor_BindTime::set_UnifyFunctions(Cor_RuleInfo *rInfo,
            Cor_Unify_Funct **unifiers,
            int *predOrders )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int     i, j, k, m, vn, arity;
    Cor_BitVector  *vars_seen;    // within a single arglist

      // Make sure the analysis has been done already
    if (!first_bind)
        return(0);

    vars_seen = new Cor_BitVector(num_vars);

#ifdef    BIND_VERBOSE
    cout << "-----------------" << endl;
    cout << *(rInfo->rule) << endl;
#endif

    for (j=0; j < num_lits; j++) {

        k = (predOrders) ? predOrders[j] : j;
        vars_seen->clear();
        Cor_ArgList *args = &(rInfo->arg_list[k]);
        arity = args->arity();

        unifiers[k] = new Cor_Unify_Funct[arity];
        for (i=0; i < arity; i++)
            unifiers[k][i] = COR_USE_UNIFY;

#ifdef    BIND_VERBOSE
        cout << "Args are: " << *args << endl;
        cout << "PredOrder is: " << k << endl; 
        cout << "unifiers[k] is: " << unifiers[k] << endl; 
#endif

        m = 0;    // Arg to check
        FOR_EACH_ARG(a, (*args)) {

            unifiers[k][m] = COR_USE_UNIFY;   // The default

            if (a->is_var()) {
                vn = a->var_num();

                if ((first_bind[vn] == j) && !vars_seen->test(vn))
                    unifiers[k][m] = COR_USE_ASSIGN;
                else
                    unifiers[k][m] = COR_USE_EQ_CHK;

                vars_seen->set(vn);

            } else if (a->isConstant()) {
                unifiers[k][m] = COR_USE_EQ_CHK;
            } else {
                // Contains variables (but not at top-level)
                unifiers[k][m] = COR_USE_MATCH;
                check_this_arg(a, vars_seen);
            }

            m++;

        } END_EACH_ARG

#ifdef    BIND_VERBOSE
        cout << "Functions are: " << endl;
        for (m=0; m < arity; m++) {
            switch (unifiers[k][m]) {

            case COR_USE_ASSIGN:
                cout << m << ": Assign.\n";
                break;

            case COR_USE_EQ_CHK:
                cout << m << ": Equal Check.\n";
                break;

            case COR_USE_MATCH:
                cout << m << ": Match.\n";
                break;

            case COR_USE_UNIFY:
                cout << m << ": Unify.\n";
                break;

            default:
                cout << m << ": Bad value.\n";
                break;
            }
        }
#endif
    }

    delete vars_seen;

    return(1);
}

// *************************************************************************
int Cor_BindTime::FixOrder(int *predOrders)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!predOrders || (num_lits < 2))
        return(0);

    // SPF - Don't try yet
    // The following doesn't work real good, and has theoretical problems
    return(1);

    int j, k, bind_ok;
    int  bind_try = 0;

    do {

        //cout << "In bind check loop\n" << flush << endl;
        bind_try++;

        bind_ok = 1;

          // Check that arithmetic predicates are not binding variables
        for (k = 0; k < num_vars; k++) {
            if ((lit_kind[first_bind[k]] == COR_BT_LIT_ARITHPRED)
                && (first_bind[k] != last_seen[k])) {
                bind_ok = 0;
                break;
            }
        }

          // Its fine
        if (bind_ok)
            break;

          // Try to fix the problem
        for (j = first_bind[k]+1; j < num_lits; j++) {
            if (lit_vars[j]->test(k) && (lit_kind[j] != COR_BT_LIT_ARITHPRED)) {
                int from = first_bind[k];
                int to   = j;

                //cout << "Attempting to fix order, moving predicate "
                //     << from << " to spot " << to << endl;

                  // Slide the literals down
                Cor_BitVector *sv_vars = lit_vars[from];
                Cor_BT_Literal_Kind sv_kind = lit_kind[from];
                int sv_pred = predOrders[from];

                for (k=from; k < to; k++) {
                    lit_vars[k] = lit_vars[k+1]; 
                    lit_kind[k] = lit_kind[k+1]; 
                    predOrders[k] = predOrders[k+1];
                }

                lit_vars[to] = sv_vars;
                lit_kind[to] = sv_kind; 
                predOrders[to] = sv_pred;
            }
        }

          //  Redo the binding-analysis.
        for (k = 0; k < num_vars; k++) {
            first_bind[k] = num_lits;
            last_seen[k]  = -1;
        }

        for (j=0; j < num_lits; j++) {
            for (k = 0; k < num_vars; k++) {
                if (lit_vars[j]->test(k)) {
                    if (j < first_bind[k])
                        first_bind[k] = j;
                    last_seen[k] = j;
                }
            }
        }

    } while (!bind_ok && (bind_try<100));

    if (bind_try >= 100) {
        //cout << "Infinite binding loop.\n";
        show_raw_info();
    }

    return(1);
}

// *************************************************************************
//  Never backtrack to relations that can succeed only once. Examples of such
//  relations are most builtins. Also, do not backtrack to a negated literal.
//  In both these cases, go back to the previous literal.
//  bt is the current backtrack position chosen.
// *************************************************************************
int Cor_BindTime::skipSingleSuccess(int bt)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // Skip single-success predicates
    while (bt >= 0) {
        if (lit_kind[bt] == COR_BT_LIT_SINGLE)
            bt--;
        else
            break;
    }

    return(bt);
}

// *************************************************************************
//  If a cut has been crossed in the process of backtracking, then the
//  backtrack point must be -1.
//      bt = current backtrack position chosen
//      src = position for which bt is the backtrack position
// *************************************************************************
int Cor_BindTime::CutCrossed(int bt, int src)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    if (bt != -1) {
        for (j=bt; j < src; j++) {
            if (lit_kind[j] == COR_BT_LIT_CUT)
                return(-1);
        }
    }

    return(bt);
}

// *************************************************************************
// Get first-failure backtrack point 
// (Used when literal fails the first time it is reached.)
//
// get-first-failure backtracking is invoked when a rhs literal
// fails the first time it is reached. In this case, execution should
// backtrack to the rightmost point that generates fresh bindings for
// it. Example : p(X,Y,C) :- q(X, Z), r(Z,A,B), s(Z,C), t(A,B,C).
// if s(Z,C) fails the first time it is reached, execution should
// backtrack to q(X,Z), not to r(Z,A,B). However, if the failure
// occurs after s(Z,C) has returned succesfully a few times, it is
// WRONG to backtrack to q(X,Z). r(Z,A,B) could generate some new
// bindings for A and B that might result in an answer tuple.
// *************************************************************************
int Cor_BindTime::get_firstFailure_BT_Pt(int *bt_array)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int m, j, k, bt;

    for (m=0; m < num_lits; m++) {

        bt = -1;
        for (j=0; j < num_vars; j++) {
            if (lit_vars[m]->test(j)) {
                k = first_bind[j];
                if ((k > bt) && (k < m)) {
                    if (lit_kind[k] == COR_BT_LIT_SINGLE) {
                        k = bt_array[k];
                        if ((k > bt) && (k < m))
                            bt = k;
                    } else 
                        bt = k;
                }
            }
        }

        //  cout << "first BT for literal " << m << " is "<< bt << endl;
        bt_array[m] = CutCrossed(bt, m);
    }

    return(0);
}

// *************************************************************************
// Get next-failure backtrack point
// (Used when literal fails after it has succeeded a few times.)
//
// get-next-failure backtracking : is invoked when a rhs literal
// fails after it has been successful at least once. 
// In this case, execution should backtrack to the rightmost point that
// generates fresh bindings for the literal or any literal after it in the
// rule execution order.
// Example : p(X,Y,C) :- q(X, Z), r(A,B), s(Z,C), t(A,B,C).
// if s(Z,C) fails when it is reached on a get-next, execution should
// backtrack to r(A,B), not to q(X,Z)).
// *************************************************************************
int Cor_BindTime::get_nextFailure_BT_Pt(int *bt_array)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int m, bt;

    for (m=0; m < num_lits; m++) {

        bt = -1;
#if    0
        int j, k;
    
        for (j=0; j < num_vars; j++) {
    
            if (last_seen[j] >= m) {
                k = first_bind[j];
                if ((k > bt) && (k < m)
                            && (lit_kind[k] != COR_BT_LIT_SINGLE))
                    bt = k;
            }
        }

  //cout << "next Backtrack for literal " << m << " is " << bt << endl;
#else
        bt = skipSingleSuccess(m - 1);
#endif

        bt_array[m] = CutCrossed(bt,m);
    }

    return(0);
}

// *************************************************************************
// Get success backtrack point
// (Used after entire rule has been sucessfully instantiated, to
//  attempt to derive another fact from it.)
//
//  Want to backtrack to the right-most literal that generates
//  new bindings for the head literal.
//
//  p(X,Y) :- q(X,Z), r(A,B), s(Z,C), t(A,B).
//
//  After the rule has suceeded, we would backtrack to q(X,Z),
//  which can result in a new binding for X.  As before,
//  if the head is allowing duplicates, we would backtrack to
//  only the previous literal, t(A,B).
// *************************************************************************
int Cor_BindTime::get_Success_BT_Pt()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int bt = -1;

#if    0
    int j, k;

    // This analysis doesn't work for:
    // ? F=foo(X,Y), call(F), print(F).
    //
    // Since F is bound "tighter" at the 2nd literal, but this
    // backtrack would proceed to the 1st literal, which would
    // fail, missing tuples (assuming foo(X,Y) had more than
    // one match).

    for (j=0; j < num_vars; j++) {

        if (head_vars->test(j)) {
            k = first_bind[j];
            if ((k > bt) && (lit_kind[k] != COR_BT_LIT_SINGLE))
                bt = k;
        }
    }

  //cout << "Success Backtrack is " << bt << endl;

#else
    bt = skipSingleSuccess(num_lits-1);
#endif

    return(CutCrossed(bt,num_lits));
}

// *************************************************************************
// For debugging (output goes to cout)
void Cor_BindTime::show_raw_info(int *predOrders)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int  j, po = 0;

    cout << "\nBinding-Time Info:\n";

    cout << "Number of Literals: " << num_lits << endl;
    cout << "Number of Unique Variables: " << num_vars << endl;

    cout << "---------------------\n";
    cout << "Literal types:\n";
    for (j=0; j < num_lits; j++) {
        if (predOrders)
            po = predOrders[j];
        switch (lit_kind[j]) {

        case COR_BT_LIT_NORMAL:
            cout << j<<": Normal. " << po << endl;
            break;
        case COR_BT_LIT_CUT:
            cout << j<<": Cut. " << po << endl;
            break;
        case COR_BT_LIT_SINGLE:
            cout << j<<": Single-success. " << po << endl;
            break;
        case COR_BT_LIT_ARITHPRED:
            cout << j<<": Arithmetic Predicate (=, >, <=, etc). " << po << endl;
            break;
        default:
            cout << j<<": Garbage value. " << po << endl;
            break;
        }
    }
    cout << "---------------------\n";

    for (j=0; j < num_vars; j++) {
        cout << " X_"<<j<<"_:  Bound @: " << first_bind[j]
             << "   Last used @: " << last_seen[j] << endl;
    }
    
    cout << "---------------------\n";
}

// *************************************************************************
