
//**************************************************************************
//    PvmConnectArray.C
//
//    The PvmConnectArray Class - Used to implement connections
//    by PVM.
//
//    A ConnectArray holds an array of Connections, of which each
//    holds an array of Users.
//     
//**************************************************************************

#ifdef    DO_PVM

#include "Identify.h"
#include "Globals.h"
#include "ClientProtocol.h"
#include "Assert.h"
#include "PvmConnectArray.h"

//**************************************************************************
//    PvmConnectArray Constructor -
//    Start the server and register it in PVM_COMM_FILE
//
//**************************************************************************

Cor_PvmConnectArray::Cor_PvmConnectArray()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    num_conn = 0;
    cur_conn = 0;
    max_conn = MAX_CONNECTIONS;
    
    for (int i = 0; i < max_conn; i++)
        connArray[i] = (Cor_Connection *) NULL;

    server_tid = -1;
    if ((server_tid = pvm_mytid()) < 0) {
        pvm_perror("pvm_mytid");
        return;
    }

    // register the server
    ofstream server_file(PVM_COMM_FILE, ios::app);
    if (!server_file) {
        cerr << "error: cannot open file " << PVM_COMM_FILE << endl;
        return;
    }
        
    server_file << "server " << server_tid << endl;
    server_file.close();
    
#ifdef    DO_EXPLAIN
//     explain_tid = -1;
//     if ((explain_tid = pvm_mytid()) < 0) {
//         pvm_perror("pvm_mytid");
//         return;
//     }

#endif

    // This next message is helpful for the test suite programs.
    cout << "[" << getpid() << "]: Coral Pvm Server started with tid " << server_tid << endl;
    cout.flush();

}

//**************************************************************************
//    PvmConnectArray Destructor -
//    Closes each connection in the array, then shutdown the server.
//    
//**************************************************************************

Cor_PvmConnectArray::~Cor_PvmConnectArray()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Close all connections
    for (int i = 0; i < MAX_CONNECTIONS; i++)
        if (connArray[i])
            delete connArray[i];

    // exit pvm
    if (pvm_exit() < 0) {
        pvm_perror("pvm_exit");
        return;
    }

    // cancel the registration in PVM_COMM_FILE
    ifstream server_file(PVM_COMM_FILE);
    if (!server_file) {
        cerr << "error: cannot open file " << PVM_COMM_FILE << endl;
        return;
    }

    char * tmp_file_name = new char[200];
    strcpy(tmp_file_name, PVM_COMM_FILE);
    strcat(tmp_file_name, "_tmp_");
    ofstream tmp_file(tmp_file_name, ios::out);
    if (!tmp_file) {
        cerr << "error: cannot open file " << tmp_file_name << endl;
        return;
    }
    
    char name[100];
    int tid;
    server_file >> name;
    while (!server_file.eof()) {
        if (!strcmp(name, "server")) {
            server_file >> tid;
            if (server_tid == tid) {
                server_file >> name;
                continue;
            }
        }
        else {
            cerr << "error: wrong format in file " << PVM_COMM_FILE << endl;
            return;
        }
        tmp_file << name << " " << tid << endl;
        
        server_file >> name;
    }

    server_file.close();
    tmp_file.close();

    if (!rename(tmp_file_name, PVM_COMM_FILE)) 
        perror("rename");
    
}

//**************************************************************************
// NextConn() -
//    This function finds the next available connection with data pending
//    and switches to it.  This function also checks for connection
//    requests on the server and accepts them. If more than one
//    connection has data waiting to be read, repeatedly calling this
//    function will switch between them in a FCFS order. 
//
//    Return: The return value is 1 if everthing is OK, 
//        or < 0 if an error occurs.
//
//**************************************************************************

int Cor_PvmConnectArray::NextConn()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int bufid, c_tid;

    while (1) {
        // busy waiting is used for the time being.
        // TODO: get a blocking version of pvm_probe
        do {
            if ((bufid = pvm_probe(-1, CORAL_TAG)) < 0) {
                pvm_perror("pvm_probe");
                return -1;
            }
        }while (bufid == 0);
    
        if (pvm_bufinfo(bufid, (int*)0, (int*)0, &c_tid) < 0) {
            pvm_perror("pvm_bufinfo");
            return -1;
        }

        for (int i = 0; i < num_conn; i++) {
            if (c_tid == client_tids[i] ) { // found data in old connection
                connArray[cur_conn = i] -> useConnection();
                return 1;
            }
        }

        // new connection request
    
        if (server_tid < 0) return -1;

        if (num_conn >= max_conn) { // refuse the connection
            // receive the request first to avoid blocking further requests
            pvm_recv(c_tid, CORAL_TAG);
            char mess[] = "Connection refused by Coral Server.\n";
            Cor_pvmbuf * pbuf = new Cor_pvmbuf(server_tid, c_tid);
            CoralStream * cs = new CoralStream(pbuf);
            *cs << mess;
            *cs << endl << CP_EXIT << endl;
            cs->flush();
            delete cs;
        }
        else {  // accept the new connection
            Cor_pvmbuf * pbuf = new Cor_pvmbuf(server_tid, c_tid);
            CoralStream * cs = new CoralStream(pbuf);
            connArray[num_conn] = new Cor_Connection(cs);
            connArray[num_conn]->conn_id = num_conn;
            client_tids[num_conn] = c_tid;
            cur_conn = num_conn;
            num_conn ++;
            connArray[cur_conn]->useConnection();
            return 1;
        }
    }

}

//**************************************************************************
// DeleteConnection() -
//    This function deletes a connection from the PvmConnectArray object.
//
//**************************************************************************

void Cor_PvmConnectArray::DeleteConnection(int conn_indx)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!connArray[conn_indx] || conn_indx >= num_conn)
        return;

    delete connArray[conn_indx];

    for (int i = conn_indx; i < num_conn-1; i++) {
        connArray[i] = connArray[i+1];
        connArray[i]->conn_id = i;
        client_tids[i] = client_tids[i+1];
    }
    connArray[num_conn - 1] = (Cor_Connection *) NULL;        
    
    num_conn --;

}

//**************************************************************************
// ExplainConnection() -
//    This function is callled when a connection request is recieved by the
//    explain server.  This function decides if the connection attempt is
//    valid and accepts or refuses it.
//
//    Return: The return value is zero if everything is okay
//        or < 0 if an error occurs or the connection is refused.
//
//**************************************************************************

int Cor_PvmConnectArray::ExplainConnection()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_EXPLAIN
    return 0;

#endif    // DO_EXPLAIN
}

//**************************************************************************

#endif    // DO_PVM
