// *************************************************************************
//	CRdbSolvers.C
// *************************************************************************

#include "CRdbManager.h"
#include "Util.h"

#ifdef DO_RDB

#include <string.h>
#include <strstream.h>

#include "CRdbSolvers.h"
#include "CRdbRelation.h"
#include "TupleIterator.h"
#include "SymTable.h"
#include "Database.h"
#include "Term.h"
#include "CoralError.h"
#include "Rdb.h"
#include "Log.h"
#include "Debug.h"


// *************************************************************************

// Standard error handling for solvers
#undef  ERR
#define ERR(A,B) { \
	CORAL_error((A),(B),fnName); \
	iterator.set_no_match(); \
	return iterator.bindenv; \
}

// *************************************************************************

Cor_BindEnv *RdbMapSolver(Cor_BuiltinRelation& , Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *fnName = "RdbMapSolver";
    char *usage = "rdb_map(coral_dbname, coral_relname(relation("
      "col1,...,colN)),...)";
    int arg_count = iterator.arg_list.count();
    if ( arg_count < 2 )
      ERR(COR_BAD_ARITY, usage);

    Cor_Term dbNameTerm (iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(dbNameTerm);
    if ( !dbNameTerm.expr.is_symbol() )
      ERR(COR_BAD_ARGLIST, usage);
    char *dbName = dbNameTerm.expr.symbol();

    // find the database
    RdbDatabase *pRdbDatabase = pCRdbManager->LookupDatabase(dbName);
    if ( pRdbDatabase == NULL )
      ERR(COR_BAD_ARGLIST, "coral rdb database not open!");

    // foreach relation to be mapped
    for( int i = 1 ; i < arg_count ; i++ ) {
	// extract the relation name 
	Cor_Term coralNameTerm(iterator.arg_list[i], iterator.bindenv);
	Cor_deep_dereference_term(coralNameTerm);
	if (! coralNameTerm.expr.is_functor() )
	  ERR(COR_BAD_ARGLIST, usage);
	Cor_CArg *coralFunc = &(coralNameTerm.expr);
	Cor_CArg coralNameArg = Cor_CArg(coralFunc->functor());
    char *coralName = coralNameArg.symbol(); 
	if ((coralFunc->args().arity() != 1) 
	    || !coralFunc->args()[0].is_functor() )
	  ERR(COR_BAD_ARGLIST, usage);
	Cor_CArg *relFunc = &(coralFunc->args()[0]);
    Cor_CArg relNameArg = Cor_CArg(relFunc->functor());
	char *relName = relNameArg.symbol();

	// extract the attributes
	RdbColNameList cols;
	int arity = relFunc->args().arity();
	for( int j = 0; j < arity; j++ ) {
	    Cor_Term colTerm ((relFunc->args())[j], iterator.bindenv);
	    Cor_deep_dereference_term(colTerm);
        if (colTerm.expr.is_symbol()) {
		    char *colName = colTerm.expr.symbol();
		    cols.Append(String(colName));
		} else { 
		    ERR(COR_BAD_ARGLIST, usage);
	    }
	}

	// create the relation
	Cor_CArg new_name = Cor_CArg(coralName);

    if (Cor_CurDB->RelationTable->lookup(new_name)) // table name in use, error
	  ERR(COR_DUP_SCHEMA,"relation NOT replaced.");

	RdbRelation *rdbRel = new RdbRelation(*pRdbDatabase, String(relName),
					      cols);
	if ( rdbRel->GetError() ) {
	    delete rdbRel;
	    // should put in new COR_RDB errors...
	    char msg[1024];
	    ostrstream oss(msg,sizeof(msg)); //replace by stream error messages
	    oss << pRdbDatabase->SystemName() << "/"
	      << pRdbDatabase->DatabaseName() << ": " 
		<< pRdbDatabase->GetErrorMessage();
	    ERR(COR_RDB_ERR, msg);
	}
	RdbRelList relList;
	relList.Append(rdbRel);
	RdbJoinCondList joinCondList;
	RdbJoin join(relList,joinCondList);
	int rc = CreateCRdbRelation(new_name,join);
	if ( rc )
	  ERR(rc, "could not create relation");
    } // for each relation to be mapped
    return iterator.bindenv ;
}

// *************************************************************************

Cor_BindEnv *RdbCommitSolver(Cor_BuiltinRelation& , Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *fnName = "RdbCommitSolver";
    char *usage = "rdb_commit([database])";
    int arg_count = iterator.arg_list.count();
    if ( arg_count != 0 && arg_count != 1 )
      ERR(COR_BAD_ARITY, usage);

    if ( arg_count == 0 ) {
    	pCRdbManager->Commit();
    } else {
	    Cor_Term dbNameTerm (iterator.arg_list[0], iterator.bindenv);
	    Cor_deep_dereference_term(dbNameTerm);
    	if ( ! dbNameTerm.expr.is_symbol() )
    	    ERR(COR_BAD_ARGLIST, usage);
    	char *dbName = dbNameTerm.expr.symbol();

    	// find the rdbms system
    	RdbDatabase *db = pCRdbManager->LookupDatabase(dbName);
    	if ( db == NULL )
    	    ERR(COR_BAD_ARGLIST, "database not found!");

    	db->Commit();
    }
    return iterator.bindenv;
}

// *************************************************************************

Cor_BindEnv *RdbRollbackSolver(Cor_BuiltinRelation&,
            Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *fnName = "RdbRollbackSolver";
    char *usage = "rdb_rollback([database])";
    int arg_count = iterator.arg_list.count();
    if ( arg_count != 0 && arg_count != 1 )
      ERR(COR_BAD_ARITY, usage);

    // make this general...
    if ( arg_count == 0 ) {
	    pCRdbManager->Rollback();
    } else {
    	Cor_Term dbNameTerm (iterator.arg_list[0], iterator.bindenv);
    	Cor_deep_dereference_term(dbNameTerm);
    	if ( !dbNameTerm.expr.is_symbol() )
    	  ERR(COR_BAD_ARGLIST, usage);
    	char *dbName = dbNameTerm.expr.symbol();

    	// find the rdbms system
    	RdbDatabase *db = pCRdbManager->LookupDatabase(dbName);
    	if ( db == NULL )
    	    ERR(COR_BAD_ARGLIST, "database not found!");

    	db->Rollback();
    }
    return iterator.bindenv;
}

// *************************************************************************

Cor_BindEnv *RdbExecuteSolver(Cor_BuiltinRelation& ,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *fnName = "RdbExecuteSolver";
    char *usage = "rdb_execute(database,command)";
    int arg_count = iterator.arg_list.count();
    if ( arg_count != 2 )
      ERR(COR_BAD_ARITY, usage);

    Cor_Term dbNameTerm (iterator.arg_list[0], iterator.bindenv);
    Cor_Term commandTerm (iterator.arg_list[1], iterator.bindenv);
    Cor_deep_dereference_term(dbNameTerm);
    Cor_deep_dereference_term(commandTerm);
    if ( ! dbNameTerm.expr.is_symbol() 
	        || ! commandTerm.expr.is_symbol() )
        ERR(COR_BAD_ARGLIST, usage);

    char *dbName = dbNameTerm.expr.symbol();
    char *cp = commandTerm.expr.symbol();
    String command = cp;

    RdbDatabase *db = pCRdbManager->LookupDatabase(dbName);
    if ( db == NULL )
      ERR(COR_BAD_ARGLIST, "database not found!");

    int rc = db->ImmediateSql(command);
    if ( rc ) {
	ostrstream oss;
	oss << db->SystemName() << '/' << db->DatabaseName() << ": " 
	  << db->GetErrorMessage();
	char *msg = oss.str();
	ERR(COR_RDB_ERR, msg);
	delete [] msg;
    }
    return iterator.bindenv;
}

// *************************************************************************

Cor_BindEnv * RdbJoinSolver(Cor_BuiltinRelation& , Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *usage = "rdb_join(newrel(rdbrel1(X,...),rdbrel2(X,...),...),...)";
    char *fnName = "RdbJoinSolver";
    int arg_count = iterator.arg_list.count();

    // for each new (join result) relation being defined
    for( int newRel = 0 ; newRel < arg_count ; newRel++ ) {
    	Cor_Term tNewRel(iterator.arg_list[newRel], iterator.bindenv);
    	Cor_deep_dereference_term(tNewRel);
    	if (!tNewRel.expr.is_functor() ) {
    	    CORAL_error(COR_BAD_ARGLIST,usage,fnName);
    	    continue;
    	}
    	
    	Cor_CArg *pfaNewRel = &(tNewRel.expr);
    	int rc = CreateCRdbJoinRelation(pfaNewRel,iterator.bindenv);
    	if ( rc ) {
    	    CORAL_error(rc,usage,fnName);
    	    continue;
    	}
    }
    return iterator.bindenv;
}

// *************************************************************************

Cor_BindEnv *RdbLoggingSolver(Cor_BuiltinRelation& ,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *usage = "rdb_logging(on|off)";
    char *fnName = "RdbLoggingSolver";

#ifndef DO_LOGGING

    ERR(COR_BAD_ARGLIST, "Not compiled with logging!");

#else // DO_LOGGING

    int arg_count = iterator.arg_list.count();
    if ( arg_count != 1)
      ERR(COR_BAD_ARGLIST, usage);

    Cor_Term tOpt(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(tOpt);
    if ( !tOpt.expr->is_symbol() )
        ERR(COR_BAD_ARGLIST, usage);

    char *Cor_opt = tOpt.expr.symbol();
    if ( strcmp(Cor_opt,"on") == 0 )
      klog.Enable();
    else if ( strcmp(Cor_opt,"off") == 0 )
      klog.Disable();
    else
      ERR(COR_BAD_ARGLIST, usage);

#endif // DO_LOGGING

    return iterator.bindenv;
}

// *************************************************************************

Cor_BindEnv * RdbOpenDBSolver(Cor_BuiltinRelation& , Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    enum {			// the position of each argument
	    WS_NAME, SYSTEM, DATABASE, USERID, PASSWORD, SERVER, 
    	MAX_ARGC, REQ_ARGC = DATABASE+1, 
    };
    char *fnName = "RdbOpenDBSolver";
    char *usage = "rdb_open_db(ws_name, system, database, "
      "[, userid[, password [, server]]])";
    int argc = iterator.arg_list.count();
    if ( argc < REQ_ARGC || argc > MAX_ARGC )
      ERR(COR_BAD_ARITY, usage);

    // extract parameters
    String argv[MAX_ARGC];	// each parameter
    for( int i = 0 ; i < argc ; i++ ) {
    	Cor_Term term(iterator.arg_list[i], iterator.bindenv);
    	Cor_deep_dereference_term(term);
    	if ( !term.expr.is_symbol() )
    	    ERR(COR_BAD_ARGLIST, usage);
    	char *cp = term.expr.symbol();
    	argv[i] = cp;
    }

    // Check if the coral db name exists
    if ( pCRdbManager->LookupDatabase(argv[WS_NAME]) != NULL )
      ERR(COR_BAD_ARGLIST, "coral database name is already in use!");

    // find the rdbms system
    RdbSystem *pRdbSystem = pCRdbManager->LookupSystem(
                                                (const char*)argv[SYSTEM]);
    if ( pRdbSystem == NULL )
      ERR(COR_BAD_ARGLIST, "invalid system name!");

    // find the database
    RdbDatabase *pRdbDatabase = pRdbSystem->LookupDatabase(argv[DATABASE],
							   argv[SERVER]);
    if ( pRdbDatabase == NULL )
      pRdbDatabase = pRdbSystem->OpenDatabase(argv[DATABASE], argv[USERID],
					      argv[PASSWORD], argv[SERVER]);
    if ( pRdbDatabase == NULL )
      ERR(COR_BAD_ARGLIST, "rdb database could not be opened!");

    // eventually, I'm going to make this into an actual workspace...
    // create...DB(EnterSymbol(argv[WS_NAME]);

    pCRdbManager->StoreDatabase(argv[WS_NAME], pRdbDatabase);
    
    return iterator.bindenv;
}


// *************************************************************************
// *************************************************************************

#endif // DO_RDB


