// *************************************************************************
//    PipelinedDerivedRelation.C
// *************************************************************************

#include <stdio.h>
#include <iomanip.h>

#include "PipelinedDerivedRelation.h"

#include "QFPipelinedModuleData.h"
#include "Compile.h"
#include "TupleIterator.h"
#include "Globals.h"
#include "Execute.h"
#include "TableStack.h"
#include "Debug.h"


// *************************************************************************

static void initRulePointers(Cor_Table *F, Cor_QFModuleData *md);

// *************************************************************************
Cor_Relation *Cor_makePipelinedRelation(Cor_QFPipelinedModuleData *md,
    Cor_CArg& name, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Relation *rel = new Cor_PipelinedDerivedRelation(md,name,arity);
    return(rel);
}

// *************************************************************************
Cor_PipelinedDerivedRelation::Cor_PipelinedDerivedRelation(
    Cor_QFPipelinedModuleData *module, Cor_CArg& _name, int new_arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    md = module;
    name = _name;
    _arity = new_arity;
    methods = NULL;
    r_kind = COR_R_DERIVED;
    has_release = 1;
    pred_offset = Cor_get_offset(module->F, name, _arity);
}

// *************************************************************************
Cor_PipelinedDerivedRelation::~Cor_PipelinedDerivedRelation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete md;
}

// *************************************************************************
void Cor_PipelinedDerivedRelation::release(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    // get rid of the state of the iterator
    if (iterator.pipe) {
        iterator.pipe->stackmark0.pop_to();
        delete iterator.pipe;
        iterator.pipe = NULL;
    }
}

// *************************************************************************
Cor_BindEnv *Cor_PipelinedDerivedRelation::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Is this method ever called?

    this->get_next_tuple(iterator);
    if (iterator.no_match()) {
        return(NULL);
    }
    return(iterator.bindenv);
}

// *************************************************************************
#define FAIL_RETURN \
  { \
    if (Cor_CurrModuleInfo != iterator.prev_module_info) { \
        Cor_CurrModuleInfo = iterator.prev_module_info; \
        Cor_T_Stack->pop(); \
    } \
    iterator.prev_module_info = NULL; \
    return NULL; \
  } 

Cor_Tuple *Cor_PipelinedDerivedRelation::get_next_tuple(
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_BindEnv *env = iterator.bindenv;
    iterator.tuple_env = NULL;
    Cor_ArgList *tempArg;
    Cor_ArgList  newArg;

#    ifdef DO_PROFILE
        // profiling info
        (Cor_GlobalProfileInfo.getnexts)++;
#    endif

    // Check if this is the first time that this iterator has been called
    // If so then perform initialization
    if (!iterator.pipe) {
#        ifdef DO_TRACE
             if (Cor_exEnv->dbg_get_next) {
               *(Cor_exEnv->trace_strm)
                   << "PipelinedDerivedRelation: new get_next called on "
                   << name.symbol() << "(";
               iterator.arg_list.print(iterator.bindenv, Cor_exEnv->trace_strm);
               *(Cor_exEnv->trace_strm) << ")\n";
             }
#        endif

        if (Cor_exEnv->C_interrupt_raised) {
            iterator.set_no_match();
            return(NULL);
        }

          // if this is the first exec on the module, 
          // we have to set up the Table
        if (!md->TableSetUp) {
            md->TableSetUp = 1;
            initRulePointers(md->F,md);
        }

        //  Since the Table is being shared by all invocations on this
        //  module, we can change it directly instead of having to 
        //  make copies of it for each execution
        //  ::: this comment is invalid :: PRAVEEN
        //  Since we would like to maintain an actual stack that
        //  reflects the module invocations, we need to make copies
        //  of the Table.

        // This is the first time get_next is being called on this iterator
        // Time to set up PipelinedExecInfo

#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_get_next) {
                *(Cor_exEnv->trace_strm)
                    << "PipelinedDerivedRelation: starting on "
                    << name.symbol() << "(";
                iterator.arg_list.print(iterator.bindenv,
                        Cor_exEnv->trace_strm);
                *(Cor_exEnv->trace_strm) << ")\n";
            }
#        endif

        iterator.pipe = new Cor_PipelinedExecInfo(md, pred_offset);
        if (!iterator.pipe->curRule) {
            // if there is no rule defining the predicate
            iterator.set_no_match();
            delete iterator.pipe;
            iterator.pipe = NULL;
#            ifdef DO_TRACE    
                if (Cor_exEnv->dbg_get_next) {
                  *(Cor_exEnv->trace_strm)
                    << "PipelinedDerivedRelation: new get_next_tuple failed on "
                    << name.symbol() << "(";
                  iterator.arg_list.print(iterator.bindenv,
                          Cor_exEnv->trace_strm);
                  *(Cor_exEnv->trace_strm) << ")\n";
                }
#            endif
            return(NULL);
        }

        // unify the arg_lists of the iterator and the head of the rule
        // keep trying all the rules till either unification succeeds or 
        // we have exhausted all the rules that define this relation
    
        // get a mark so that we can undo the changes if it fails
        iterator.pipe->stackmark0.get_mark();
        // make an ArgList*
        // that points to the ArgList for the rule head
    
        tempArg = (iterator.pipe)->rInfo->arg_list;
    
        // sanity check
        ASSERT(iterator.pipe->rInfo->num_literals ==
            iterator.pipe->curRule->num_preds);
        ASSERT(env != NULL);

        newArg = (tempArg[(iterator.pipe)->rInfo->num_literals]);

        while (Cor_unify_arg_lists(newArg,iterator.arg_list,iterator.pipe->env,
            iterator.bindenv) != COR_U_SUCCEED)
        {
            // unification failed - backtrack
            iterator.pipe->stackmark0.pop_to();
            // Try the next rule
            if (!(iterator.pipe->NextRule(md))) {
                // if we have exhausted all the rules, give up
                iterator.set_no_match();

                //iterator.pipe->stackmark0.pop_to();
                delete iterator.pipe;
                iterator.pipe = NULL;
#                ifdef DO_TRACE    
                    if (Cor_exEnv->dbg_get_next) {
                        *(Cor_exEnv->trace_strm)
                          << "PipelinedDerivedRelation: new get_next failed on "
                          << name.symbol() << "(";
                        iterator.arg_list.print(iterator.bindenv,
                            Cor_exEnv->trace_strm);
                        *(Cor_exEnv->trace_strm) << ")\n";
                    }
#                endif
                return NULL;
                //FAIL_RETURN;
            }

            tempArg = (iterator.pipe)->rInfo->arg_list;
            newArg = tempArg[(iterator.pipe)->rInfo->num_literals];
        }
    } else {
         //iterator.pipe->stackmark0.pop_to();
#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_get_next) {
                *(Cor_exEnv->trace_strm)
                  << "PipelinedDerivedRelation: get_next_tuple repeat call on "
                  << name.symbol() << "(";
                iterator.arg_list.print(iterator.bindenv,
                        Cor_exEnv->trace_strm);
                *(Cor_exEnv->trace_strm) << ")\n";
            }
#        endif
    }

    // check to see that an interrupt has not been raised
    if (Cor_exEnv->C_interrupt_raised) {
#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_get_next) {
                *(Cor_exEnv->trace_strm)
                  << "PipelinedDerivedRelation: get_next failed on "
                  << name.symbol() << "(";
                iterator.arg_list.print(iterator.bindenv,
                    Cor_exEnv->trace_strm);
                *(Cor_exEnv->trace_strm) << ") due to interrupt\n";
            }
#        endif
        iterator.set_no_match();
        //iterator.pipe->stackmark0.pop_to();
        delete iterator.pipe;
        iterator.pipe = NULL;
        return NULL;
        //FAIL_RETURN;
    }
    
    iterator.prev_module_info = Cor_CurrModuleInfo;
    if (Cor_CurrModuleInfo != md->module_info) {
        Cor_T_Stack->push(md->F);
        Cor_CurrModuleInfo = md->module_info;
    }

    // Now go ahead and call Solve on PipelinedExecInfo
    // keep trying as long as no match is found
    while (!((iterator.pipe)->Solve())) {
        iterator.pipe->stackmark0.pop_to();
        // try the next rule 
        if ( ((iterator.pipe)->cut_crossed) || !(iterator.pipe->NextRule(md)))
        {
            // if we have exhausted all the rules,
            // or have failed because one rule tried to backtrack
            // past a cut, give up
            iterator.set_no_match();
            delete iterator.pipe;
            iterator.pipe = NULL;

#            ifdef DO_TRACE
                if (Cor_exEnv->dbg_get_next) {
                    *(Cor_exEnv->trace_strm)
                      << "PipelinedDerivedRelation: get_next failed on "
                      << name.symbol() << "(";
                    iterator.arg_list.print(iterator.bindenv,
                        Cor_exEnv->trace_strm);
                    *(Cor_exEnv->trace_strm) << ")\n";
                }
#            endif
            //return NULL;
            FAIL_RETURN;
        }
        tempArg = (iterator.pipe)->rInfo->arg_list;
        newArg = (tempArg[(iterator.pipe)->rInfo->num_literals]);

        // See if the head of this rule unifies with the iterator
        while (Cor_unify_arg_lists( newArg, iterator.arg_list,
            (iterator.pipe)->env, env) != COR_U_SUCCEED)
        {
            // unification failed - backtrack
            iterator.pipe->stackmark0.pop_to();

            // Try the next rule
            if (!(iterator.pipe->NextRule(md))) {
                // if we have exhausted all the rules, give up
                iterator.set_no_match();
                //iterator.pipe->stackmark0.pop_to();
                delete iterator.pipe;
                iterator.pipe = NULL;
#                ifdef DO_TRACE
                    if (Cor_exEnv->dbg_get_next) {
                        *(Cor_exEnv->trace_strm)
                          << "PipelinedDerivedRelation: get_next failed on "
                          << name.symbol() << "(";
                        iterator.arg_list.print(iterator.bindenv,
                            Cor_exEnv->trace_strm);
                        *(Cor_exEnv->trace_strm) << ")\n";
                    }
#                endif
                //return NULL;
                FAIL_RETURN;
            }

            tempArg = (iterator.pipe)->rInfo->arg_list;
            newArg = tempArg[(iterator.pipe)->rInfo->num_literals];
        }
    }

    // Success
    iterator.reset_no_match();

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm)
            << "PipelinedDerivedRelation: get_next_tuple succeeded on "
            << name.symbol() << "(";
        iterator.arg_list.print(iterator.bindenv, Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << ")\n";
    }
#endif
    
    //return NULL;
    FAIL_RETURN;
}

#undef FAIL_RETURN

// *************************************************************************
void Cor_PipelinedDerivedRelation::print_name(CoralStream *outf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *outf << setw(15) << name.symbol() << "/" << arity()
          << "    : (pipelined) \n";
}

// *************************************************************************
static void initRulePointers(Cor_Table *F, Cor_QFModuleData *md)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ModulePred *predicates = md->predicates->next;
    F->moduledata = md;
    
    for (; predicates; predicates = predicates->next) {
        Cor_CArg name = predicates->pred->name;
        int arity = predicates->pred->arity;
        int offset = predicates->offset;

        switch (predicates->kind) {
            Cor_Relation *pipe_rel;

        case QueryPredKind:
            F->rels[offset].AssignRel(
                    Cor_Find_External_Relation(name, arity));
            break;

        case LocalPredKind:
            // We create PipelinedDerivedRelations for local
            // predicates only. No tuples will be actually inserted into any of 
            // these Relations, but we still need it so as to allow us to
            // do a get_next() on it

            pipe_rel = Cor_makePipelinedRelation(
                    (Cor_QFPipelinedModuleData *) md,name,arity);
            pipe_rel->set_localpiped();
            F->rels[offset].AssignRel(pipe_rel);
            break;

        case ExternalPredKind:
            F->rels[offset].AssignRel(
                Cor_Find_External_Relation(name, arity));
            
            Cor_AnnoStruct *cur;
            for (cur= predicates->index_annos; cur; cur=cur->next) {
                if (cur->anno->arglist2) {
                     F->rels[offset].relPtr->add_index(cur->anno->arglist1,
                        cur->num_vars, cur->anno->arglist2);
                } else {
                    F->rels[offset].relPtr->add_index(cur->anno->arglist1) ;
                }
            }
            
            F->rels[offset].relPtr->check_subsum = predicates->check_subsum;
             
            break;

         case MagicQueryPredKind:
         default:
            *(Cor_exEnv->err_strm) << "This cannot happen.\n";
            break;
         }
    }
}

// *************************************************************************

