// *************************************************************************
//    Catalog.C
// *************************************************************************

#ifdef WITH_PERSISTENCE

#include "Globals.h"
#include "CoralError.h"
#include "SMUtil.h"
#include "sm_client.h"
#include "Config.h"
#include "Schema.h"
#include "IndexStruct.h"
#include "BPlusTreeIndex.h"
#include "FileDescStruct.h"
#include "BufferGroup.h"
#include "Transaction.h"
#include "Catalog.h"
#include "StorageManager.h"
#include "PersistentRelation.h"
#include "Debug.h"

// *************************************************************************
Cor_Catalog::~Cor_Catalog()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete prel;
}

// *************************************************************************
/*
    Initializes catalog by creating catalog file and catalog index if
    they do not exist already.  
*/

int Cor_Catalog::init(char *cat_file_name, char *cat_index_name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int data_length = -1;
    int catalog_exists = 1;
    int e;

    // Check if catalog file already exists
    e = sm_GetRootEntry(SM.volid(), cat_file_name, (char *)&cat_fid,
        &data_length);

    if ((sm_errno == esmBADROOTNAME) || (data_length != sizeof(FID))) {
        // Entry does not exist
        catalog_exists = 0;
        e = sm_CreateFile(SM.bufGroup(), SM.volid(), &cat_fid);
        if (ErrorCheck(e, "sm_CreateFile")) {
            SM.abortTrans();
            return -1;
        }

        e = sm_SetRootEntry(SM.volid(), cat_file_name, (char *)&cat_fid,
            sizeof(FID));
        if (ErrorCheck(e, "sm_SetRootEntry")) {
            SM.abortTrans();
            return -1;
        }
    }

    data_length = -1;
    // Check if catalog index already exists
    e = sm_GetRootEntry(SM.volid(), cat_index_name, (char *)&cat_iid,
         &data_length);

    if (sm_errno == esmBADROOTNAME) {
        // Entry does not exist
        if (catalog_exists) {
            *(Cor_exEnv->err_strm) << 
                "Catalog file exists, but index does not exist !\n";
            *(Cor_exEnv->err_strm) << "Disk has to be reformatted !\n";
            SM.abortTrans();
            return -1;
        }

        e = sm_CreateIndex( SM.volid(), SM.bufGroup(), SM_BTREENDX, SM_string,
            COR_MAX_REL_NAME_LEN, sizeof(OID), TRUE, &cat_iid);
        if (ErrorCheck(e, "sm_CreateIndex")) {
            SM.abortTrans();
            return -1; 
        }

        e = sm_SetRootEntry(SM.volid(), cat_index_name, (char *)&cat_iid,
            sizeof(IID));
        if (ErrorCheck(e, "sm_SetRootEntry")) {
            SM.abortTrans();
            return -1;
        }
    }
    else if (!catalog_exists) {
        *(Cor_exEnv->err_strm) << 
            "Catalog index exists, but file does not exist !\n";
        *(Cor_exEnv->err_strm) << "Disk has to be reformatted !\n";
        SM.abortTrans();
        return -1;
     }
     else if (data_length != sizeof(IID)) {
        SM.abortTrans();
        *(Cor_exEnv->err_strm) << "Error creating catalog file\n";
        return -1;
     }

    // now create main-memory objects corresponding to each persistent relation

    SMCURSOR cursor;
    KEY key1;

    char value_buff1[COR_MAX_KEY_LEN+1];

    for (int i = 0; i < COR_MAX_KEY_LEN ; i++)
        value_buff1[i] = '\0';
 
    key1.length = (TWO)(COR_MAX_KEY_LEN);
    key1.valuePtr = (void *)(value_buff1);

    // Initialize the index scan
    e = sm_FetchInit(&cat_iid, SM.bufGroup(), SM_BOF, SM_EOF, &cursor);
        
    ErrorCheck(e, "sm_FetchInit");
    for (;;) {
        OID oid;
        char eof;

        // Get the next entry from the index
        key1.length = (TWO)(COR_MAX_KEY_LEN);
        e = sm_FetchNext(&cursor, &key1, (void *)(&oid), &eof);
        ErrorCheck(e, "sm_FetchNext");

        // If there are no more index entries, return.
        if (eof) {
            SM.commitTrans();
            return 0;
        }

        if (e < 0) {
#            ifdef DEBUG
                *(Cor_exEnv->err_strm) << 
                    "CORAL::WARNING Entry with key length too large\n";
#            endif
            ErrorCheck(e,"sm_FetchNext");
            SM.abortTrans();
            return 1;
        }

        char name[1000];
        int arity;
        ((char *)key1.valuePtr)[key1.length] = '\0';
        sscanf(((char *)key1.valuePtr), "%s%3d", name, &arity);

        int error;
          // create a persistent relation structure (LEAK)
        prel = new Cor_PersistentRel(Cor_CArg(name), arity, error, &oid);
    }

    if (!SM.commitTrans()) {
        *(Cor_exEnv->err_strm) << "Commit error\n";
        SM.abortTrans();
        return -1;
    }

    return 0;
}

// *************************************************************************
/*
    Checks catalog index to see if a file of the specified name
    exists. If so, the OID of the entry for that file in the catalog file
    is returned via the var parameter. A return value of 0 implies that
    the entry was not found.
*/

int Cor_Catalog::lookupFile(char *name, OID *oid)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;

    SMCURSOR cursor;
    KEY key;
    key.length = (TWO)(strlen(name)+1);
    key.valuePtr = (void *)name;
    char eof;

    int e = sm_FetchInit(&cat_iid, SM.bufGroup(), &key, SM_EQ, &key, SM_EQ,
        &cursor);
    if (ErrorCheck(e, "sm_FetchInit")) return 0;

    e = sm_FetchNext(&cursor, NULL, (void *)oid, &eof);
    if (ErrorCheck(e, "sm_FetchNext")) return 0;

    if (eof) return 0 ; // Entry not found

    return 1;
}

// *************************************************************************
/*
    Adds a file to the catalog file, and adds the appropriate entry in the
    catalog index. The data to be entered is pointed to by obj.
*/

int Cor_Catalog::addFile(char *file_name, Cor_FileDescStruct *obj)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;

    // Add the entry to the catalog file
    OID oid;

    int e = sm_CreateObject(SM.bufGroup(), &cat_fid, NEAR_LAST_PHYSICAL, NULL,
        NULL, (FOUR)sizeof(Cor_FileDescStruct), (char *)obj, &oid);
    if (ErrorCheck(e, "sm_CreateObject")) return 0;

    // Add the oid to the catalog index
    KEY key;
    key.length = (TWO)(strlen(file_name)+1);
    key.valuePtr = (void *)file_name;

    e = sm_InsertEntry(&cat_iid, SM.bufGroup(), &key, &oid);
    if (ErrorCheck(e, "sm_InsertEntry")) return 0;

    return 1;
}

// *************************************************************************
//    Deletes a file from the catalogs.

int Cor_Catalog::deleteFile(char *file_name, OID *oid)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;

    KEY key;
    key.valuePtr = (void *)file_name;
    key.length = (TWO)(strlen(file_name) + 1);

    int e = sm_RemoveEntry(&cat_iid, SM.bufGroup(), &key, oid);
    if (ErrorCheck(e, "sm_RemoveEntry")) return 0;

    // Delete the file descriptor object itself
    e = sm_DestroyObject(SM.bufGroup(), oid);
    if (ErrorCheck(e, "sm_DestroyObject")) return 0;

    return 1;
}

// *************************************************************************
void Cor_Catalog::listFiles(ostream *outf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    SMCURSOR cursor;
    KEY key1;

    char value_buff1[COR_MAX_KEY_LEN+1];

    for (int i = 0; i < COR_MAX_KEY_LEN ; i++) {
        value_buff1[i] = '\0';
    }

    key1.length = (TWO)(COR_MAX_KEY_LEN);
    key1.valuePtr = (void *)(value_buff1);

    // Initialize the index scan
    int e = sm_FetchInit(&cat_iid, SM.bufGroup(), SM_BOF, SM_EOF, &cursor);
        
    ErrorCheck(e, "sm_FetchInit");
    for (;;) {
        OID oid;
        char eof;

        // Get the next entry from the index
        key1.length = (TWO)(COR_MAX_KEY_LEN);
        e = sm_FetchNext(&cursor, &key1, (void *)(&oid), &eof);

        // If there are no more index entries, return.
        if (eof) return;

        if (e < 0) {
#            ifdef DEBUG
                *(Cor_exEnv->err_strm) << 
                    "CORAL::WARNING Entry with key length too large\n";
#            endif
            ErrorCheck(e,"sm_FetchNext");
            return;
        }

        USERDESC * u_desc;
        if (! SM.retrieveObject(&oid, &u_desc)) {
            CORAL_error(COR_ESM_ERR, NULL, "Catalog::listFiles");
            return;
        }
        
        value_buff1[key1.length] = '\0';
        *outf << value_buff1 << " : ";
        ((Cor_FileDescStruct *)(u_desc->basePtr))->schema.printon(outf);
        *outf  << " : (exodus)\n";
//*outf << ": count " << ((FileDescStruct *)(u_desc->basePtr))->count << endl;
    }
}

// *************************************************************************

#endif WITH_PERSISTENCE
