// *************************************************************************
//  Main.C
//  Implements Magic, Supplementary Magic and Context Rule Rewriting
//  (as in Kemp, Ramamohanrao & Somogyi(VLDB 90))
// *************************************************************************

#include <iostream.h>

#include "Defunct.h"
#include "Linearity.h"
#include "Adorn.h"
#include "Print.h"
#include "SipOrder.h"
#include "SupMagic.h"
#include "FullMagic.h"
#include "RuleFactor.h"

#include "CoralInit.h"
#include "Collection.h"
#include "ModuleInfo.h"
#include "Scc.h"
#include "Unparse.h"
#include "RelationMisc.h"
#include "Compile.h"
#include "ParseUtils.h"
#include "BasePreds.h"
#include "Globals.h"
#include "ExportInfo.h"
#include "Annotations.h"
#include "Util.h"
#include "Debug.h"


// *************************************************************************
//
//  Notes:
//  1) What do I do about opt_share_functors and debugmagic ??
//  2) Factoring and SupMagicIndexing do not co-exist. If both options are set,
//      supplementary magic with supmagicindexing is done. Factoring is not.
//
// *************************************************************************

extern void preprocess_for_grouping(Cor_Collection *orig_rules, 
                    Cor_Collection *pre_ad_rules);

// *************************************************************************

CoralStream *inf  = NULL;
CoralStream *outf = NULL;

int    optAdorn=0, optFactor=0, optDefunct=0;

// Used as extern by funct/defunct code
int    orig_rule_count = 0;

Cor_Collection *pre_ad_rules;
Cor_Collection *factor_rules;
Cor_Collection *magic_rules;

// *************************************************************************

extern int show_prompt;
extern Cor_Scc *Cor_SccList;

// *************************************************************************

extern "C" int yyparse();

// *************************************************************************

static void printUsage(char *arg0);
static void mark_scc(Cor_Scc *scc, LinearKind kind);
static void check_ord_search(Cor_ParserStruct *Cor_parserStruct);
static void find_status();
static void generate_index_annos(Cor_Collection *rules,
    Cor_PredAnnotations*& index_annos);
static void clear_sips(Cor_Collection *rules);

void do_factor_rewriting();
void add_magic_rule_labels (Cor_Collection *factor_rules);

// *************************************************************************
int main(int ac, char **av)
{
  int i;
  char *infile = "stdin", *outfile = NULL;

#ifdef    DO_CALL_STACK
    callstk = new CallStk();
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

  init_coral(av[0]);
  inf  = cin_strm;
  outf = cout_strm;

  pre_ad_rules = new Cor_Collection();
  factor_rules = new Cor_Collection();
  magic_rules  = new Cor_Collection();

  for (i=1; i<ac && (*av[i]=='-'); ++i) {
      if (strcmp(av[i]+1, "a") == 0)
          optAdorn = 1;
      else if (strcmp(av[i]+1, "d") == 0)
          optDefunct = 1;
      else if (strcmp(av[i]+1, "f") == 0)
          optFactor = 1;
      else if (strcmp(av[i]+1, "i") == 0)
          infile = av[++i];
      else if (strcmp(av[i]+1, "o") == 0)
          outfile = av[++i];
      else if (strcmp(av[i]+1, "factor") == 0) {
          if (++i < ac)
              Cor_exEnv->C_use_factoring_default = atoi(av[i]);
      }
      else if (strcmp(av[i]+1, "magic") == 0) {
          if (++i < ac)
              Cor_exEnv->C_use_magic_default = atoi(av[i]);
      }
      else if (strcmp(av[i]+1, "supmagic") == 0) {
          if (++i < ac)
              Cor_exEnv->C_use_supplementary_magic_default = atoi(av[i]);
      }
      else if (strcmp(av[i]+1, "pipelining") == 0) {
          if (++i < ac)
              Cor_exEnv->C_use_pipelining_default = atoi(av[i]);
      }
      else if (strcmp(av[i]+1, "ordsearch") == 0) {
          if (++i < ac)
              Cor_exEnv->C_use_ordsearch_default = atoi(av[i]);
      }
      else {
          printUsage(av[0]);
          exit(1);
      }
  }

  if (strcmp(infile,"stdin") != 0)
      if ((inf = Cor_open_infile(infile)) == NULL){
         *cerr_strm << "Could not open file " << infile << " for input.\n";
         exit(1);
      }
  Cor_parseStack->handle_new_file(inf, infile);
    

  if (outfile){
      if ((outf = Cor_open_outfile(outfile)) == NULL){
            *cerr_strm << "Could not open file " << outfile
                << " for output. (using stdout)\n";
         outf = cout_strm;
      }
  }

  Cor_exEnv->C_print_prompt_default = 0;
  Cor_exEnv->C_exec_mode = COR_DOING_FACTOR;

  for(;;) {

      if (yyparse()) {
          cerr_strm->puts("syntax error -- quitting\n");
          break;
      }

      if (Cor_parserStruct->cur_collect->count()) {
          if (Cor_exEnv->C_at_end_module)
              do_factor_rewriting();
          else {
              cerr_strm->puts("CORAL::factor: Warning - end_module missing\n");
              do_factor_rewriting();
          }
      } else
        break;
  }

  delete pre_ad_rules;
  delete factor_rules;
  delete magic_rules;

  exit_coral();

#ifdef    DO_CALL_STACK
    delete callstk;
    callstk = NULL;
#endif

  return(0);
}

// *************************************************************************
static void printUsage(char *arg0)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
   cerr << "\nUsage: "
        << arg0 << " [-l] [-a] [-c] [-r] [-i infile] [-o outfile]\n\n";
   cerr << "    -a    Print the program with initial adornments.\n";
   cerr << "    -d    Print the Extended Datalog program.\n";
   cerr << "    -f    Print the program with factoring status of rules.\n";
   cerr << "    -i    Use infile instead of stdin.\n";
   cerr << "    -o    Use outfile instead of stdout.\n\n";
}

// *************************************************************************
static void mark_scc(Cor_Scc *scc, LinearKind kind)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (Cor_Predicate *pred = scc->preds; pred != NULL; pred = pred->next) {
        Cor_Clause *clause = pred->clauses;
        for (; clause != *pred->clause_tail; clause = clause->pnext) {
            if (!clause->seen) continue;
            clause->factor_status = (int)kind;
        }
    }
}

// *************************************************************************
/*
    This is used to determine whether the module needs to be computed using
    Ordered Search.  There are two cases:
        (1) If there is aggregation in the head, and a derived literal in the
            body.
        (2) If there is genuine negation in the body of the rule, this should
            work for both magic and supmagic.  I'm not sure about supmagic
            -- Praveen
*/
// *************************************************************************

static void check_ord_search(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_parserStruct->CurModule->UseOrdSearch)
        return;
    
    Cor_Clause *adrule;
    int j, nrules;

    nrules = Cor_parserStruct->cur_collect->count();

    for (j=0; j < nrules; j++) {
        adrule = Cor_parserStruct->cur_collect->get_ith(j);

        for (int i=0; i < adrule->num_preds; ++i) {
            if (!adrule->preds[i]->base_pred) {
                // check if there is aggregation in the head
                if (adrule->head->args.has_grouping_args()) {
                    COR_set_use_ordsearch(*(Cor_parserStruct->CurModule), 1);
                    return;
                }
                // check if there is genuine negation in the head of the rule
                if (adrule->preds[i]->negated) {
                    // negated and derived predicates
                    COR_set_use_ordsearch(*(Cor_parserStruct->CurModule), 1);
                    return;
                }
            }
        }
    }
}


// *************************************************************************
/*
    Assigns linear status to the rules that define the predicates reachable
    from the exported predicates. For each exported predicate, an SCC analysis 
    is performed, and the rules in each scc are analysed independently.
    Factoring can be performed at any level of such stratified rules.
    For those sccs that cannot be factored, magic transformation is done.

    PRAVEEN: What happens if there is more than one such exported pred?
    Answer: There is a fixed scc structure for the rules, irrespective of
    the exported predicates. So the same rule cannot be given one linear
    status for one exported predicate, and a different status for another.
*/
// *************************************************************************

static void find_status()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    for (Cor_ExportInfo *export_info =
        Cor_parserStruct->CurModule->ExportedPreds;
        export_info != NULL; export_info = export_info->next)
    {
        Cor_CArg query_symbol;
        Cor_CArg exported_name = export_info->pred_name;
        int query_arity;

        if (export_info->adorn_symbol == 0) {
            query_arity = 0;
            query_symbol = exported_name;
        }
        else {
            query_arity = (int) export_info->adorn_symbol.symlen();
            query_symbol = Cor_CArg(exported_name.symbol());
        }

        export_info->predicate = Cor_LookupPredicate(Cor_parserStruct,
                query_symbol, query_arity);
        if (!export_info->predicate) {
            *cerr_strm << "Error: Exported predicate "
                    << query_symbol.symbol() << " not defined in module\n";
            continue;
        }

        // Do Scc (Strongly Connected Components) analysis.
        Cor_Clause *clause;
        int j, nrules;

        nrules = Cor_parserStruct->cur_collect->count();

        for (j=0; j < nrules; j++) {
            clause = Cor_parserStruct->cur_collect->get_ith(j);

            clause->seen = 0;

            clause->recursive = 0;
            for (int ipred = 0; ipred < clause->num_preds; ipred++) {
                clause->preds[ipred]->predicate->work = NULL;
                clause->preds[ipred]->predicate->mark = 0;
            }
        }

        Cor_initialize_scc_structs();

        Cor_find_predicates_invoked_from(export_info->predicate);
        Cor_number_sccs(Cor_SccList);

        if (Cor_parserStruct->CurModule->NestedSccEval)
          continue;

        for (Cor_Scc *scc = Cor_SccList; scc != NULL; scc = scc->next) {
            Cor_Predicate *pred; Cor_Clause *clause; int ipred;

            // Use the scc->recursive field to specify whether the scc
            // contains more than one recursive predicate.

            // If the module is not to be factored, do magic instead.
            if (!Cor_parserStruct->CurModule->UseFactoring)
            {
                mark_scc(scc, COR_NON_LINEAR);
                continue;
            }

            scc->recursive = 0;

            for (pred = scc->preds; pred != NULL; pred = pred->next) {
                clause = pred->clauses;
                for (; clause != *pred->clause_tail; clause = clause->pnext) {
                    if (!clause->seen) continue;
                    for (ipred = 1; ipred <= clause->num_preds; ipred++) {
                        Cor_Predicate *child =
                            clause->preds[ipred-1]->predicate;
                        if (child->scc == scc) clause->recursive = 1;
                        if ((child->scc == scc) && (child != pred)
                            && (!clause->preds[ipred-1]->base_pred))
                        {
                            // Mutually recursive scc !!
                            scc->recursive = 1;
                            break;
                        }
                        if (scc->recursive) break;
                    }
                    if (scc->recursive) break;
                }
                if (scc->recursive) break;
              }

            if (scc->recursive) {
                mark_scc(scc, COR_NON_LINEAR);
                continue;
            }

            for (pred = scc->preds; pred != NULL; pred = pred->next) {
                if (pred->scc == scc) break;
            }

            // At this stage, pred contains the only derived predicate
            // of the scc.
            if (pred == NULL) {
                cerr_strm->puts("SOMETHING WRONG HERE !\n");
                break;
            }

            clause = pred->clauses;
            for (; clause != *pred->clause_tail; clause = clause->pnext) {
                if (!clause->seen) continue;
                if (!assign_linear_status(clause)) {
                      // Use recursive field to signify that this scc
                      // is not factorable.
                      scc->recursive = 1;
                      break;
                }
                else if (clause->factor_status == PSEUDO_LT_LINEAR) {
                    // As pointed out by David Kemp, pseudo-lt-linear rules
                    // are not compatible with other kinds of linear rules.
                    // Also, in the case of pseudo-lt-linear rules and
                    // lt-linear rules, context rewriting is only as
                    // efficient as magic, so might as well do magic
                    // on it.
                      scc->recursive = 1;
                    break;
                }
            }
            if (scc->recursive) mark_scc(scc, COR_NON_LINEAR);
        }
    }
}

// *************************************************************************
static void generate_index_annos(Cor_Collection *rules,
    Cor_PredAnnotations*& index_annos)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char buf[COR_MAX_ARGLIST_SIZE+1];
    int i, p;
    Cor_ArgList *al1;
    Cor_CArg     tmp_arg;
    Cor_Clause  *rule;
    int j, nrules;
    Cor_BitVector bv_tmp;

    nrules = rules->count();

    for (j=0; j < nrules; j++) {
        rule = rules->get_ith(j);

        for (p=0; p < rule->num_preds; ++p) {
            if ((rule->preds[p]->base_pred) && (rule->preds[p]->adorn.length)) {
                for (i = 0; i < rule->preds[p]->adorn.length; ++i) {
                    buf[i] = rule->preds[p]->adorn[i] ? 'b' : 'f';
                }
                buf[i] = '\0';
                tmp_arg = Cor_CArg(buf);
                al1 = new Cor_ArgList(1);
                (*al1)[0] = tmp_arg;
                Cor_add_make_index_annotation(index_annos,
                    rule->preds[p]->name(), &bv_tmp, al1, NULL);
            }
        }
    }
}


// *************************************************************************
static void clear_sips(Cor_Collection *rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause  *rule;
    int j, nrules;

    nrules = rules->count();

    for (j=0; j < nrules; j++) {
        rule = rules->get_ith(j);
        rule->sips = NULL;
    }
}

// *************************************************************************
void do_factor_rewriting() 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    TermList **terms; // For factoring

    if (Cor_parserStruct->CurModule->UsePipelining) {
        // If the module uses pipelining, there is really no need to do any
        // rewriting.  However, in order to correctly generate indices, there
        // must be an analysis of how bindings are propogated.  One way to
        // do this is to first mark all preds as derived preds and perform
        // the adornment phase.  This will cause the base preds to be
        // adorned in all the possible ways.  The actual base preds can now
        // be marked, and index annotations generated for just these preds.
        // Finally, these index annotations are dumped along with the
        // original module.

        Cor_parserStruct->CurModule->PreProcessing = 0;
        mark_all_as_derived(Cor_parserStruct->cur_collect);
        pre_ad_rules->empty();
        pre_ad_rules->use_hash_table(1);

        if (!adorn(Cor_parserStruct->cur_collect, pre_ad_rules)) {
            cerr_strm->puts("Unable to proceed with magic/factor rewriting!\n");
            exit(1);
        }

        if (optAdorn) {
            cerr_strm->puts("Printing adorned rules\n");
            print_factor(pre_ad_rules, cerr_strm, 0);
            cerr_strm->puts("\n\n");
        }
        
        mark_basepreds(pre_ad_rules);
        generate_index_annos(pre_ad_rules,
                Cor_parserStruct->make_index_annotations);
        clear_sips(Cor_parserStruct->cur_collect);
        print_factor(Cor_parserStruct->cur_collect, outf, 0);
        return;
    }

    if (!Cor_parserStruct->CurModule->Rewriting ||
        (!Cor_parserStruct->CurModule->UseMagic &&
            !Cor_parserStruct->CurModule->UseSupplementaryMagic))
    {
        *cerr_strm << "Warning: Magic rewriting not performed for Module "
            << Cor_parserStruct->CurModule->name.symbol() << ".\n";
        Cor_parserStruct->CurModule->PreProcessing = 0;
        print_factor(Cor_parserStruct->cur_collect, outf, 0);
        return;
    }

    mark_basepreds(Cor_parserStruct->cur_collect);

    pre_ad_rules->empty();
    pre_ad_rules->use_hash_table(1);

    // For monotonic programs, there is no need to do ordered search.
    // Also, if ordered search is already set, don't bother checking again.
    if (!Cor_parserStruct->CurModule->UseOrdSearch &&
        !Cor_parserStruct->CurModule->Monotonic)
            check_ord_search(Cor_parserStruct);

    if (Cor_parserStruct->CurModule->UseFactoring &&
        Cor_parserStruct->CurModule->UseOrdSearch)
    {
        *(Cor_exEnv->err_strm) << "Ordered search is turned on.\n";
        *(Cor_exEnv->err_strm) << "Have to turn factoring off: sorry.\n";
        COR_set_use_factoring(*(Cor_parserStruct->CurModule), 0);     
    }

    preprocess_for_grouping(Cor_parserStruct->cur_collect, pre_ad_rules);

    if (optAdorn) {
        *cerr_strm << "Printing pre-adorned rules\n";
        print_factor(pre_ad_rules, cerr_strm, 0);
        *cerr_strm << "\n\n";
    }

    factor_rules->empty();
    factor_rules->use_hash_table(1);

    if (!adorn(pre_ad_rules, factor_rules)) {
        *cerr_strm << "Unable to proceed with magic/factor rewriting!\n";
        exit(1);
    }

    if (optAdorn) {
        *cerr_strm << "Printing adorned rules\n";
        print_factor(factor_rules, cerr_strm, 0);
        *cerr_strm << "\n\n";
    }

    magic_rules->empty();

    Cor_Clause *r;
    int j, nrules;
    
    nrules = factor_rules->count();

    for (j=0; j < nrules; j++) {
        r = factor_rules->get_ith(j);
        r->head->predicate->clauses = NULL;
        r->head->predicate->clause_tail = &(r->head->predicate->clauses);
    }

    for (j=0; j < nrules; j++) {
        r = factor_rules->get_ith(j);
        Cor_AddClauseToPredicate(r->head->predicate, r);
    }

    if (Cor_parserStruct->CurModule->UseFactoring) {
        // NOTE:  It is important that the TermList be created only at
        // this point, where the size of the adorned rule collection is
        // known.
        orig_rule_count = factor_rules->count();
        terms = new TermList*[orig_rule_count];

        defunctify(factor_rules, terms); 
        if (optDefunct) {
            *cerr_strm << "Printing rules in Extended Datalog form\n";
            print_factor(factor_rules, cerr_strm, 0);
            *cerr_strm << "\n\n";
        }

        // Do an SCC analysis, and assign linear status to each rule.
        find_status();

        if (Cor_parserStruct->CurModule->UseFactoring) {
            functify(factor_rules, terms); 
        }

        if (optFactor) {
            *cerr_strm << "Printing rules after factoring\n";
            print_factor(factor_rules, cerr_strm, 1);
            *cerr_strm << "\n\n";
        }

        factor_rewrite(factor_rules, magic_rules);
    } else {
        siporder(factor_rules);

        if (Cor_parserStruct->CurModule->NestedSccEval) {
          // Do Scc analysis
          find_status(); 

            // classify rules
          Cor_Clause *rule;
          int j, nrules;

          nrules = factor_rules->count();

          for (j=0; j < nrules; j++) {
            rule = factor_rules->get_ith(j);

            int sup = 0;
            for (int i=0; i < rule->num_preds; ++i) {
              if (!rule->preds[i]->base_pred &&
                    rule->preds[i]->predicate->scc->scc_no != 
                    rule->head->predicate->scc->scc_no) 
                sup++;
            }

            if (sup)
              rule->supp = 1; // should be rewritten using sup_magic
            else
              rule->supp = 0; // should be rewritten using magic
            
          }
          // perform rewriting for nested-scc evaluation
          supmagic(factor_rules, magic_rules);
          fullmagic(factor_rules, magic_rules);
        }
        else if (Cor_parserStruct->CurModule->UseSupplementaryMagic) {
            supmagic(factor_rules, magic_rules);
            // supopt();
        }
        else if ( Cor_parserStruct->CurModule->UseMagic)
            fullmagic(factor_rules, magic_rules);
    }
    // Set @no_preprocessing to show that factoring/magic has been performed
    Cor_parserStruct->CurModule->PreProcessing = 0;

    add_magic_rule_labels(magic_rules);

    print_factor(magic_rules, outf, 0);
}

// *************************************************************************
void add_magic_rule_labels (Cor_Collection *factor_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // find the number of original rules
    int tot_rules = 1;
    int tmp_rule_number = 0;

    Cor_Clause *factorrule;
    int j, nrules;

    nrules = factor_rules->count();

    for (j=0; j < nrules; j++) {
      factorrule = factor_rules->get_ith(j);

      if (sscanf(factorrule->clause_names->ith_symbol(0), "%d",
            &tmp_rule_number) == 1)
         if (tmp_rule_number > tot_rules)
            tot_rules = tmp_rule_number;
    }

    int *magic_rule_label_counter = new int[tot_rules+1];

    for (int i=1; i<=tot_rules; i++)
      magic_rule_label_counter[i] = 1;

    for (j=0; j < nrules; j++) {
      factorrule = factor_rules->get_ith(j);

      int org_rule_number;

      if (!(factorrule->head->kind == COR_MAGIC || 
                    factorrule->head->kind == COR_SUPMAGIC)) {
             // original rule
            ASSERT(! factorrule->clause_names->is_empty());
            factorrule->clause_names->add("0",0);

      } else {

        ASSERT(! factorrule->clause_names->is_empty());

          // derived rule
        sscanf(factorrule->clause_names->ith_symbol(0), "%d", 
               &org_rule_number);

        factorrule->clause_names->empty();

        factorrule->clause_names->add(org_rule_number,0);

        factorrule->clause_names->add(
                    magic_rule_label_counter[org_rule_number],0);        

        magic_rule_label_counter[org_rule_number]++;
      }
    }
 
    delete [] magic_rule_label_counter;
}

// *************************************************************************
