// *************************************************************************
//    OutputExport.C
// *************************************************************************

#include <stdio.h>
#include <string.h>

#include "OutputExport.h"

#include "HandleExport.h"
#include "MyString.h"
#include "HashDefines.h"
#include "Lexer.h"
#include "Main.h"

/*****************************************************************************/

static void gen_type_decl(FILE *output_file, kind *k);
static void gen_arg_assign(FILE *output_file, kind *k, int i);
static void gen_ret_assign(FILE *output_file, kind *k, char *fn_name,
    int argcount, handle_export *export_info);
static void gen_arg_ret_assign(FILE *output_file, kind *k, int argcount);
static void gen_header(FILE *output_file, char *fn_name);
static void gen_init_assign(FILE *output_file, int i);
static void gen_tail(FILE *output_file);
static void gen_class_decl(FILE*, char *exported_fn_name, char *fn_name,
    int argcount, mystring *export_dummy_fn, char *dummy_temp_name);
static void gen_fail_code(FILE *output_file);

static void arg_code(FILE*, kind *k);
static void ret_code(FILE*, kind *k);

/*****************************************************************************/
void output_init_export(char *, char *dummy_fn_name) 
{
    fprintf(output_file, "static void %s(); \n", dummy_fn_name);
}

/*****************************************************************************/
void output_export_code(handle_export *export_info,
    mystring *export_dummy_fn, char *dummy_temp_name,
    mystring * /* export_header_decl */)
{
    /*  export_header_decl->print(output_file);  */
    
    kind *k;
    char *fn_name = export_info->get_name();
    
    gen_header(output_file, fn_name);
    
    int argcount = export_info->get_num_of_args();
    int i;
    for (i = 0; i < argcount; i++) {
        k = export_info->get_arg(i);
        if (k->opt_neg_sign == 0) {
            //    Load the input args into into ouput arglist.
            gen_init_assign(output_file, i);
        }
    }
    
    for (i = 0; i < argcount; i++) {
        k = export_info->get_arg(i);
        k->argname = concat1("arg",i);
        k->init_pointer = 1;
        //    Set up the types
        arg_code(output_file, k);
        
        //    Generate the defs for the output defs.
        gen_type_decl(output_file, k);

        //    Generate assigns to the vars.
        gen_arg_assign(output_file, k, i);
    }
    
    k = export_info->get_return_type();
    k->argname = "retvalue";
    k->init_pointer = 0;
    ret_code(output_file, k);
    
    
    //    Generate the type for the retvalue
    gen_type_decl(output_file, k);

    //    Call the fn
    gen_ret_assign(output_file, k, fn_name, argcount, export_info);
    
    for (i = 0; i < argcount; i++) {
        k = export_info->get_arg(i);
        if (k->opt_neg_sign) {
            gen_arg_ret_assign(output_file, k, i);
        }
    }
    
    gen_tail(output_file);
    
    char *exported_fn_name = export_info->get_exported_name();
    if (exported_fn_name) {
        gen_class_decl(output_file, exported_fn_name, fn_name,argcount,
            export_dummy_fn, dummy_temp_name);
    }
    else {
        gen_class_decl(output_file, fn_name, fn_name,argcount,
            export_dummy_fn, dummy_temp_name);
    }

    line_directive(output_file);
}

/*****************************************************************************/
static void gen_type_decl(FILE *output_file, kind *k)
{
    //    NOTE : the only types that should have a pointer are char and Arg.
    //    Here too, only char * is valid. So the code here is special-cased
    //    for this.

    fprintf(output_file, "\t%s%c %s ; \n", k->type,
        (k->pointer && !strchr(k->type,'*') ? '*': ' '), 
        k->argname);    // gen the type
}

/*****************************************************************************/
static void gen_arg_assign(FILE *output_file, kind *k, int i)
{
    if (!(k->opt_neg_sign)) {
        if (k->init_pointer == 0) {
            if (k->pointer > 1) fprintf(output_file, "*");
            fprintf(output_file, "\t%s = ", k->argname);
            if (k->pointer) {
                if (k->type)
                    fprintf(output_file, "(%s)", concat2(k->type, " *"));
                if (k->make)
                    fprintf(output_file, "%s", k->make);
            }
            else {
                if (k->type)
                    fprintf(output_file, "(%s)", k->type);
                if (k->make)
                    fprintf(output_file, "%s", k->make);
                else
                    fprintf(output_file, "*");
            }
            fprintf(output_file, "((query->args())[%d]); \n", i);
            return;
        }
        fprintf(output_file, "\t%s = (%s)%s((query->args())[%d]); \n",
            k->argname, k->type, k->make, i);
    }
}

/*****************************************************************************/
static void gen_ret_assign(FILE *output_file, kind *k, char *fn_name,
    int argcount, handle_export *export_info)
{
    char *temp = concat1("arg",argcount-1);
    fprintf(output_file, "\t");
    if (k->init_pointer) {
        if (k->pointer)
            fprintf(output_file, "*");
    }
    fprintf(output_file, "%s = %s(", k->argname, fn_name);
    int i;
    for (i = 0; i < argcount - 1; i++)
        fprintf(output_file, "%s, ", (export_info->get_arg(i))->argname);
    fprintf(output_file, "%s);\n", temp);

    gen_fail_code(output_file);

    switch(k->value){

    case FLOAT_T:
        fprintf(output_file,
            "\t(*alist)[%d] = (double) %s;\n", argcount, k->argname);
        break;

    case DOUBLE_T:
    case INT_T:
    case SHORT_INT:
    case LONG_INT:
    case UNSIGNED_INT:
    case UNSIGNED_SHORT_INT:
    case UNSIGNED_LONG_INT:
        fprintf(output_file,
            "\t(*alist)[%d] = %s;\n", argcount, k->argname);
        break;

    default:
        if (k->make) {
            fprintf(output_file, "\t(*alist)[%d] = make_arg(%s%s);\n",
                argcount, k->make, k->argname);
        }
        else {
            fprintf(output_file,
                "\t(*alist)[%d] = make_arg(%s);\n", argcount, k->argname);
        }
        break;
    }
}

/*****************************************************************************/
static void gen_arg_ret_assign(FILE  *output_file, kind *k, int argcount)
{
    concat1("arg",argcount);

    if (k->make) {
        fprintf(output_file,
            "\t(*alist)[%d] = make_arg(%s%s);\n", argcount,k->make, k->argname);
    }
    else {
        fprintf(output_file,
            "\t(*alist)[%d] = make_arg(%s);\n", argcount,k->argname);
    }
}

/*****************************************************************************/
char *concat1(char *s, int i)
{
    char buf[10];

    sprintf(buf, "%d", i);
    int slen = strlen(s);
    int buflen = strlen(buf);
    int len = slen + buflen + 1;
    char *t = new char[len];
    strncpy(t, s, slen);
    strncpy(t+slen, buf, buflen);
    t[len-1] = '\0';
    return t;
}

/*****************************************************************************/
char *concat2(char *s1, char *s2)
{
    int s1len = strlen(s1);
    int s2len = strlen(s2);
    int len = s1len + s2len + 1;
    char *t = new char[len];
    strncpy(t, s1, s1len);
    strncpy(t+s1len, s2, s2len);
    t[len-1] = '\0';
    return t;
}

/*****************************************************************************/
//    Generates the first few lines of the solver.
  
static void gen_header(FILE *output_file, char *fn_name)
{
    char *temp = concat2(fn_name, "solver");
    
    fprintf(output_file, "static Cor_Tuple * %s(Cor_Tuple *query)\n{\n", temp);
    fprintf(output_file, "\tif (query == NULL) {\n\t    fprintf(stderr,");
    fprintf(output_file, "\"Null argument passed to %s \\n\"); \n", temp);
    fprintf(output_file, "\t    return NULL; \n\t} \n");
    fprintf(output_file, "\tint arity = query->arity(); \n");
    fprintf(output_file, "\tCor_ArgList *alist = new Cor_ArgList(arity);\n");
}

/*****************************************************************************/
static void gen_init_assign(FILE *output_file, int i)
{
    fprintf(output_file, "\t(*alist)[%d] = (query->args())[%d];\n", i,i);
}

/*****************************************************************************/
static void gen_tail(FILE *output_file)
{
    fprintf(output_file, "\tCor_Tuple *return_tuple = new Cor_Tuple(alist);\n");
    fprintf(output_file, "\treturn return_tuple;\n}\n");
}

/*****************************************************************************/
static void gen_class_decl(FILE*, char *exported_fn_name, char *fn_name,
    int argcount, mystring *export_dummy_fn, char *dummy_temp_name)
{
    char buf[256];
    sprintf(buf,
    "  %s = new Cor_BuiltinUserTupleRelation(%d, new Cor_CArg(\"%s\"), %s);\n",
        dummy_temp_name, argcount+1, exported_fn_name,
        concat2(fn_name, "solver"));
    export_dummy_fn->append(buf);
}

/*****************************************************************************/
static void gen_fail_code(FILE *output_file)
{
    fprintf(output_file,
        "\tif(%s){\n\t\t%s=0;\n\t\tdelete alist;\n\t\treturn NULL;\n\t}\n",
        CORAL_FAIL_VAR,CORAL_FAIL_VAR);
}

/*****************************************************************************/
static void arg_code(FILE*, kind *k)
{
    if (k->opt_neg_sign == 0) {
        switch(k->value) {
        case CHAR_T:
            if (k->pointer) {
                k->type = "char *";
                k->make = "make_string";
            }
            else {
                k->type = "char";
                k->make = "make_int";
            }
            k->_typedef = k->type;
            break;
        case INT_T:
            k->type = "int";
            k->make = "make_int";
            k->_typedef = k->type;
            break;
        case FLOAT_T:
            k->type = "float";
            k->make = "make_float";
            k->_typedef = k->type;
            break;
        case DOUBLE_T:
            k->type = "double";
            k->make = "make_double";
            k->_typedef = k->type;
            break;
        case SHORT_INT:
            k->type = "short";
            k->make = "make_short";
            k->_typedef = k->type;
            break;
        case LONG_INT:
            k->type = "long";
            k->make = "make_long";
            k->_typedef = k->type;
            break;
        case UNSIGNED_INT:
            k->type = "unsigned";
            k->make = "make_int";
            k->_typedef = k->type;
            break;
        case UNSIGNED_SHORT_INT:
            k->type = "unsigned short";
            k->make = "make_short";
            k->_typedef = k->type;
            break;
        case UNSIGNED_LONG_INT:
            k->type = "unsigned long";
            k->make = "make_long";
            k->_typedef = k->type;
            break;
        case CARG_T:
            k->type = "Cor_CArg *";
            k->make = "";
            //k->init_pointer = 0;
            k->_typedef = k->type;
            break;
        case STRING_T:
            k->type = "char";
            k->pointer = 1;
            k->make = "make_string";
            k->init_pointer = 0;
            k->_typedef = k->type;
            break;
        case CLASS_T:
            k->type = k->name;
            k->make = "";
            k->_typedef = k->type;
            break;
        case STRUCT_T:
            k->type = k->name;
            k->make = "";
            k->_typedef = k->type;
            break;
        default:
            k->type = NULL;
            k->make = "";
            k->_typedef = k->type;
            break;
        }
    }
    else {
        switch(k->value) {
        case CHAR_T:
            k->type = "char";
            if (k->pointer)
                k->make = "(char)*";
            else
                k->make = "(int)";
            break;
        case INT_T:
            k->type = "int";
            if (k->pointer)
                k->make = "*";
            break;
        case FLOAT_T:
            k->type = "float";
            if (k->pointer)
                k->make = "(double)*";
            else
                k->make = "(double)";
            break;
        case DOUBLE_T:
            k->type = "double";
            if (k->pointer)
                k->make = "*";
            break;
        case SHORT_INT:
            k->type = "short";
            if (k->pointer)
                k->make = "(int)*";
            else
                k->make = "(int)";
            break;
        case LONG_INT:
            k->type = "long";
            if (k->pointer)
                k->make = "(int)*";
            else
                k->make = "(int)";
            break;
        case UNSIGNED_INT:
            k->type = "unsigned";
            if (k->pointer)
                k->make = "(int)*";
            else
                k->make = "(int)";
            break;
        case UNSIGNED_SHORT_INT:
            k->type = "unsigned short";
            if (k->pointer)
                k->make = "(int)*";
            else
                k->make = "(int)";
            break;
        case UNSIGNED_LONG_INT:
            k->type = "unsigned long";
            if (k->pointer)
                k->make = "(int)*";
            else
                k->make = "(int)";
            break;
        case CARG_T:
            if (!(k->pointer)) { 
                fprintf(stderr,
                    "Warning :: -ve with Cor_CArg does not work correctly \n");
                k->make = "(Cor_CArg*)&";
            }
            else {
                if (k->pointer > 1) {
                    /*    printf("string is %s \n", k->argname); */
                    k->make = "(Cor_CArg*)*";
                }
                else {
                    k->make = "(Cor_CArg*)";
                }
            }
            k->type = "Cor_CArg";
            break;
        case STRING_T:
            k->type = "char";
            k->pointer = 1;
            k->make = NULL;
            break;
        case CLASS_T:
            k->type = k->name;
            k->make = NULL;
            break;
        case STRUCT_T:
            k->type = k->name;
            k->make = NULL;
            break;
        default:
            k->type = NULL;
            k->make = NULL;
            break;
        }
        if (!(k->pointer)) 
            fprintf(stderr,
                "Warning : -ve sign should be used with pointer args only, since C has only pass by value \n");
    }
}

/*****************************************************************************/
static void ret_code(FILE *, kind *k)
{
    switch(k->value) {
    case CHAR_T:
        k->type = "char";
        if (k->pointer)
            k->make = "(char)*";
        else
            k->make = "(int)";
        break;
    case INT_T:
        k->type = "int";
        if (k->pointer)
            k->make = "*";
        break;
    case FLOAT_T:
        k->type = "float";
        if (k->pointer)
            k->make = "(double)*";
        else
            k->make = "(double)";
        break;
    case DOUBLE_T:
        k->type = "double";
        if (k->pointer)
            k->make = "*";
        break;
    case SHORT_INT:
        k->type = "short";
        if (k->pointer)
            k->make = "(int)*";
        else
            k->make = "(int)";
        break;
    case LONG_INT:
        k->type = "long";
        if (k->pointer)
            k->make = "(int)*";
        else
            k->make = "(int)";
        break;
    case UNSIGNED_INT:
        k->type = "unsigned";
        if (k->pointer)
            k->make = "(int)*";
        else
            k->make = "(int)";
        break;
    case UNSIGNED_SHORT_INT:
        k->type = "unsigned short";
        if (k->pointer)
            k->make = "(int)*";
        else
            k->make = "(int)";
        break;
    case UNSIGNED_LONG_INT:
        k->type = "unsigned long";
        if (k->pointer)
            k->make = "(int)*";
        else
            k->make = "(int)";
        break;
    case CARG_T:
        k->type = "Cor_CArg *";
        k->make = "";
        /*
        if (!(k->pointer))
            k->init_pointer = 1;
        */
        k->pointer = 1;
        break;
    case STRING_T:
        k->type = "char";
        k->pointer = 1;
        k->make = NULL;
        break;
    case CLASS_T:
        k->type = k->name;
        k->make = NULL;
        break;
    case STRUCT_T:
        k->type = k->name;
        k->make = NULL;
        break;
    default:
        k->type = NULL;
        k->make = NULL;
        break;
    }
}

/*****************************************************************************/
void output_call_dummy(char *dummy_fn_name)
{
    fprintf(output_file, "%s();\n", dummy_fn_name);
    line_directive(output_file);
}

/*****************************************************************************/

